"""
Sample parsers.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import os

import numpy as np

import eta.core.image as etai
import eta.core.serial as etas
import eta.core.utils as etau

import fiftyone as fo
import fiftyone.core.clips as foc
import fiftyone.core.labels as fol
import fiftyone.core.metadata as fom
from fiftyone.core.sample import Sample
import fiftyone.core.utils as fou
import fiftyone.core.validation as fov
import fiftyone.utils.eta as foue

fouv = fou.lazy_import("fiftyone.utils.video")


def add_images(dataset, samples, sample_parser, tags=None):
    """Adds the given images to the dataset.

    This operation does not read the images.

    See :ref:`this guide <custom-sample-parser>` for more details about
    adding images to a dataset by defining your own
    :class:`UnlabeledImageSampleParser`.

    Args:
        dataset: a :class:`fiftyone.core.dataset.Dataset`
        samples: an iterable of samples that can be parsed by ``sample_parser``
        sample_parser: a :class:`UnlabeledImageSampleParser` instance to use to
            parse the samples
        tags (None): an optional tag or iterable of tags to attach to each
            sample

    Returns:
        a list of IDs of the samples that were added to the dataset
    """
    if not sample_parser.has_image_path:
        raise ValueError(
            "Sample parser must have `has_image_path == True` to add its "
            "samples to the dataset"
        )

    if not isinstance(sample_parser, UnlabeledImageSampleParser):
        raise ValueError(
            "`sample_parser` must be a subclass of %s; found %s"
            % (
                etau.get_class_name(UnlabeledImageSampleParser),
                etau.get_class_name(sample_parser),
            )
        )

    if etau.is_str(tags):
        tags = [tags]
    elif tags is not None:
        tags = list(tags)

    def parse_sample(sample):
        sample_parser.with_sample(sample)

        image_path = sample_parser.get_image_path()

        if sample_parser.has_image_metadata:
            metadata = sample_parser.get_image_metadata()
        else:
            metadata = None

        return Sample(filepath=image_path, metadata=metadata, tags=tags)

    try:
        num_samples = len(samples)
    except:
        num_samples = None

    _samples = map(parse_sample, samples)
    return dataset.add_samples(
        _samples, num_samples=num_samples, expand_schema=False
    )


def add_labeled_images(
    dataset,
    samples,
    sample_parser,
    label_field=None,
    tags=None,
    expand_schema=True,
    dynamic=False,
):
    """Adds the given labeled images to the dataset.

    This operation will iterate over all provided samples, but the images will
    not be read (unless the sample parser requires it in order to compute image
    metadata).

    See :ref:`this guide <custom-sample-parser>` for more details about
    adding labeled images to a dataset by defining your own
    :class:`LabeledImageSampleParser`.

    Args:
        dataset: a :class:`fiftyone.core.dataset.Dataset`
        samples: an iterable of samples that can be parsed by ``sample_parser``
        sample_parser: a :class:`LabeledImageSampleParser` instance to use to
            parse the samples
        label_field (None): controls the field(s) in which imported labels are
            stored. If the parser produces a single
            :class:`fiftyone.core.labels.Label` instance per sample, this
            argument specifies the name of the field to use; the default is
            ``"ground_truth"``. If the parser produces a dictionary of labels
            per sample, this argument can be either a string prefix to prepend
            to each label key or a dict mapping label keys to field names; the
            default in this case is to directly use the keys of the imported
            label dictionaries as field names
        tags (None): an optional tag or iterable of tags to attach to each
            sample
        expand_schema (True): whether to dynamically add new sample fields
            encountered to the dataset schema. If False, an error is raised
            if a sample's schema is not a subset of the dataset schema
        dynamic (False): whether to declare dynamic attributes of embedded
            document fields that are encountered

    Returns:
        a list of IDs of the samples that were added to the dataset
    """
    if not sample_parser.has_image_path:
        raise ValueError(
            "Sample parser must have `has_image_path == True` to add its "
            "samples to the dataset"
        )

    if not isinstance(sample_parser, LabeledImageSampleParser):
        raise ValueError(
            "`sample_parser` must be a subclass of %s; found %s"
            % (
                etau.get_class_name(LabeledImageSampleParser),
                etau.get_class_name(sample_parser),
            )
        )

    if isinstance(label_field, dict):
        label_key = lambda k: label_field.get(k, k)
    elif label_field is not None:
        label_key = lambda k: label_field + "_" + k
    else:
        label_field = "ground_truth"
        label_key = lambda k: k

    if etau.is_str(tags):
        tags = [tags]
    elif tags is not None:
        tags = list(tags)

    def parse_sample(sample):
        sample_parser.with_sample(sample)

        image_path = sample_parser.get_image_path()

        if sample_parser.has_image_metadata:
            metadata = sample_parser.get_image_metadata()
        else:
            metadata = None

        label = sample_parser.get_label()

        sample = Sample(filepath=image_path, metadata=metadata, tags=tags)

        if isinstance(label, dict):
            sample.update_fields({label_key(k): v for k, v in label.items()})
        elif label is not None:
            sample[label_field] = label

        return sample

    # Optimization: if we can deduce exactly what fields will be added during
    # import, we declare them now and set `expand_schema` to False
    try:
        can_expand_now = issubclass(sample_parser.label_cls, fol.Label)
    except:
        can_expand_now = False

    if expand_schema and can_expand_now:
        dataset._ensure_label_field(label_field, sample_parser.label_cls)
        expand_schema = False

    try:
        num_samples = len(samples)
    except:
        num_samples = None

    _samples = map(parse_sample, samples)
    return dataset.add_samples(
        _samples,
        expand_schema=expand_schema,
        dynamic=dynamic,
        num_samples=num_samples,
    )


def add_videos(dataset, samples, sample_parser, tags=None):
    """Adds the given videos to the dataset.

    This operation does not read the videos.

    See :ref:`this guide <custom-sample-parser>` for more details about
    adding videos to a dataset by defining your own
    :class:`UnlabeledVideoSampleParser`.

    Args:
        dataset: a :class:`fiftyone.core.dataset.Dataset`
        samples: an iterable of samples that can be parsed by ``sample_parser``
        sample_parser: a :class:`UnlabeledVideoSampleParser` instance to use to
            parse the samples
        tags (None): an optional tag or iterable of tags to attach to each
            sample

    Returns:
        a list of IDs of the samples that were added to the dataset
    """
    if not isinstance(sample_parser, UnlabeledVideoSampleParser):
        raise ValueError(
            "`sample_parser` must be a subclass of %s; found %s"
            % (
                etau.get_class_name(UnlabeledVideoSampleParser),
                etau.get_class_name(sample_parser),
            )
        )

    if etau.is_str(tags):
        tags = [tags]
    elif tags is not None:
        tags = list(tags)

    def parse_sample(sample):
        sample_parser.with_sample(sample)

        video_path = sample_parser.get_video_path()

        if sample_parser.has_video_metadata:
            metadata = sample_parser.get_video_metadata()
        else:
            metadata = None

        return Sample(filepath=video_path, metadata=metadata, tags=tags)

    try:
        num_samples = len(samples)
    except:
        num_samples = None

    _samples = map(parse_sample, samples)

    # @todo: skip schema expansion and set media type before adding samples
    return dataset.add_samples(
        _samples, num_samples=num_samples, expand_schema=True
    )


def add_labeled_videos(
    dataset,
    samples,
    sample_parser,
    label_field=None,
    tags=None,
    expand_schema=True,
    dynamic=False,
):
    """Adds the given labeled videos to the dataset.

    This operation will iterate over all provided samples, but the videos will
    not be read/decoded/etc.

    See :ref:`this guide <custom-sample-parser>` for more details about
    adding labeled videos to a dataset by defining your own
    :class:`LabeledVideoSampleParser`.

    Args:
        dataset: a :class:`fiftyone.core.dataset.Dataset`
        samples: an iterable of samples that can be parsed by ``sample_parser``
        sample_parser: a :class:`LabeledVideoSampleParser` instance to use to
            parse the samples
        label_field (None): controls the field(s) in which imported labels are
            stored. If the parser produces a single
            :class:`fiftyone.core.labels.Label` instance per sample/frame, this
            argument specifies the name of the field to use; the default is
            ``"ground_truth"``. If the parser produces a dictionary of labels
            per sample/frame, this argument can be either a string prefix to
            prepend to each label key or a dict mapping label keys to field
            names; the default in this case is to directly use the keys of the
            imported label dictionaries as field names
        tags (None): an optional tag or iterable of tags to attach to each
            sample
        expand_schema (True): whether to dynamically add new sample fields
            encountered to the dataset schema. If False, an error is raised
            if a sample's schema is not a subset of the dataset schema
        dynamic (False): whether to declare dynamic attributes of embedded
            document fields that are encountered

    Returns:
        a list of IDs of the samples that were added to the dataset
    """
    if not isinstance(sample_parser, LabeledVideoSampleParser):
        raise ValueError(
            "`sample_parser` must be a subclass of %s; found %s"
            % (
                etau.get_class_name(LabeledVideoSampleParser),
                etau.get_class_name(sample_parser),
            )
        )

    if isinstance(label_field, dict):
        label_key = lambda k: label_field.get(k, k)
    elif label_field is not None:
        label_key = lambda k: label_field + "_" + k
    else:
        label_field = "ground_truth"
        label_key = lambda k: k

    if etau.is_str(tags):
        tags = [tags]
    elif tags is not None:
        tags = list(tags)

    def parse_sample(sample):
        sample_parser.with_sample(sample)

        video_path = sample_parser.get_video_path()

        if sample_parser.has_video_metadata:
            metadata = sample_parser.get_video_metadata()
        else:
            metadata = None

        label = sample_parser.get_label()
        frames = sample_parser.get_frame_labels()

        sample = Sample(filepath=video_path, metadata=metadata, tags=tags)

        if isinstance(label, dict):
            sample.update_fields({label_key(k): v for k, v in label.items()})
        elif label is not None:
            sample[label_field] = label

        if frames is not None:
            frame_labels = {}

            for frame_number, _label in frames.items():
                if isinstance(_label, dict):
                    frame_labels[frame_number] = {
                        label_key(field_name): label
                        for field_name, label in _label.items()
                    }
                elif _label is not None:
                    frame_labels[frame_number] = {label_field: _label}

            sample.frames.merge(frame_labels)

        return sample

    try:
        num_samples = len(samples)
    except:
        num_samples = None

    _samples = map(parse_sample, samples)
    return dataset.add_samples(
        _samples,
        expand_schema=expand_schema,
        dynamic=dynamic,
        num_samples=num_samples,
    )


class SampleParser(object):
    """Base interface for sample parsers.

    :class:`SampleParser` instances are used to parse samples emitted by
    dataset iterators when ingesting them into
    :class:`fiftyone.core.dataset.Dataset` instances.

    The general recipe for using :class:`SampleParser` instances is as
    follows::

        sample_parser = SampleParser(...)

        for sample in samples:
            sample_parser.with_sample(sample)
            field = sample_parser.get_<field>()

    where ``field`` is a subclass specific field to parse from the sample.
    """

    def __init__(self):
        self._current_sample = None

    @property
    def current_sample(self):
        """The current sample.

        Raises:
            ValueError: if there is no current sample
        """
        if self._current_sample is None:
            raise ValueError(
                "No current sample. You must call `with_sample()` before "
                "trying to get information about a sample"
            )

        return self._current_sample

    def with_sample(self, sample):
        """Sets the current sample so that subsequent calls to methods of this
        parser will return information from the given sample.

        Guaranteed to call :meth:`clear_sample` before setting the current
        sample.

        Args:
            sample: a sample
        """
        self.clear_sample()
        self._current_sample = sample

    def clear_sample(self):
        """Clears the current sample.

        Also clears any cached sample information stored by the parser.
        """
        self._current_sample = None


class UnlabeledImageSampleParser(SampleParser):
    """Interface for :class:`SampleParser` instances that parse unlabeled image
    samples.

    Instances of this class must return images in ``numpy`` format.

    The general recipe for using :class:`UnlabeledImageSampleParser` instances
    is as follows::

        sample_parser = UnlabeledImageSampleParser(...)

        for sample in samples:
            sample_parser.with_sample(sample)
            img = sample_parser.get_image()
            if sample_parser.has_image_path:
                image_path = sample_parser.get_image_path()

            if sample_parser.has_image_metadata:
                image_metadata = sample_parser.get_image_metadata()
    """

    @property
    def has_image_path(self):
        """Whether this parser produces paths to images on disk for samples
        that it parses.
        """
        raise NotImplementedError("subclass must implement has_image_path")

    @property
    def has_image_metadata(self):
        """Whether this parser produces
        :class:`fiftyone.core.metadata.ImageMetadata` instances for samples
        that it parses.
        """
        raise NotImplementedError("subclass must implement has_image_metadata")

    def get_image(self):
        """Returns the image from the current sample.

        Returns:
            a numpy image
        """
        raise NotImplementedError("subclass must implement get_image()")

    def get_image_path(self):
        """Returns the image path for the current sample.

        Returns:
            the path to the image on disk
        """
        if not self.has_image_path:
            raise ValueError(
                "This '%s' does not provide image paths"
                % etau.get_class_name(self)
            )

        raise NotImplementedError("subclass must implement get_image_path()")

    def get_image_metadata(self):
        """Returns the image metadata for the current sample.

        Returns:
            a :class:`fiftyone.core.metadata.ImageMetadata` instance
        """
        if not self.has_image_metadata:
            raise ValueError(
                "This '%s' does not provide image metadata"
                % etau.get_class_name(self)
            )

        raise NotImplementedError(
            "subclass must implement get_image_metadata()"
        )


class UnlabeledVideoSampleParser(SampleParser):
    """Interface for :class:`SampleParser` instances that parse unlabeled video
    samples.

    The general recipe for using :class:`UnlabeledVideoSampleParser` instances
    is as follows::

        sample_parser = UnlabeledVideoSampleParser(...)

        for sample in samples:
            sample_parser.with_sample(sample)
            video_path = sample_parser.get_video_path()
            video_metadata = sample_parser.get_video_metadata()
    """

    @property
    def has_video_metadata(self):
        """Whether this parser produces
        :class:`fiftyone.core.metadata.VideoMetadata` instances for samples
        that it parses.
        """
        raise NotImplementedError("subclass must implement has_video_metadata")

    def get_video_path(self):
        """Returns the video path for the current sample.

        Returns:
            the path to the video on disk
        """
        raise NotImplementedError("subclass must implement get_video_path()")

    def get_video_metadata(self):
        """Returns the video metadata for the current sample.

        Returns:
            a :class:`fiftyone.core.metadata.VideoMetadata` instance
        """
        if not self.has_video_metadata:
            raise ValueError(
                "This '%s' does not provide video metadata"
                % etau.get_class_name(self)
            )

        raise NotImplementedError(
            "subclass must implement get_video_metadata()"
        )


class ImageSampleParser(UnlabeledImageSampleParser):
    """Sample parser that parses unlabeled image samples.

    This implementation assumes that the provided sample is either an image
    that can be converted to numpy format via ``np.asarray()`` or the path
    to an image on disk.
    """

    @property
    def has_image_path(self):
        return True

    @property
    def has_image_metadata(self):
        return False

    def get_image(self):
        image_or_path = self.current_sample
        if etau.is_str(image_or_path):
            return etai.read(image_or_path)

        return np.asarray(image_or_path)

    def get_image_path(self):
        image_or_path = self.current_sample
        if etau.is_str(image_or_path):
            return image_or_path

        raise ValueError(
            "Cannot extract image path from samples that contain images"
        )


class VideoSampleParser(UnlabeledVideoSampleParser):
    """Sample parser that parses unlabeled video samples.

    This implementation assumes that the provided sample is a path to a video
    on disk.
    """

    @property
    def has_video_metadata(self):
        return False

    def get_video_path(self):
        return self.current_sample


class LabeledImageSampleParser(SampleParser):
    """Interface for :class:`SampleParser` instances that parse labeled image
    samples.

    Instances of this class must return images in ``numpy`` format and labels
    as :class:`fiftyone.core.labels.Label` instances.

    The general recipe for using :class:`LabeledImageSampleParser` instances
    is as follows::

        sample_parser = LabeledImageSampleParser(...)

        for sample in samples:
            sample_parser.with_sample(sample)
            img = sample_parser.get_image()
            label = sample_parser.get_label()

            if sample_parser.has_image_path:
                image_path = sample_parser.get_image_path()

            if sample_parser.has_image_metadata:
                image_metadata = sample_parser.get_image_metadata()
    """

    @property
    def has_image_path(self):
        """Whether this parser produces paths to images on disk for samples
        that it parses.
        """
        raise NotImplementedError("subclass must implement has_image_path")

    @property
    def has_image_metadata(self):
        """Whether this parser produces
        :class:`fiftyone.core.metadata.ImageMetadata` instances for samples
        that it parses.
        """
        raise NotImplementedError("subclass must implement has_image_metadata")

    @property
    def label_cls(self):
        """The :class:`fiftyone.core.labels.Label` class(es) returned by this
        parser.

        This can be any of the following:

        -   a :class:`fiftyone.core.labels.Label` class. In this case, the
            parser is guaranteed to return labels of this type
        -   a list or tuple of :class:`fiftyone.core.labels.Label` classes. In
            this case, the parser can produce a single label field of any of
            these types
        -   a dict mapping keys to :class:`fiftyone.core.labels.Label` classes.
            In this case, the parser will return label dictionaries with keys
            and value-types specified by this dictionary. Not all keys need be
            present in the imported labels
        -   ``None``. In this case, the parser makes no guarantees about the
            labels that it may return
        """
        raise NotImplementedError("subclass must implement label_cls")

    def get_image(self):
        """Returns the image from the current sample.

        Returns:
            a numpy image
        """
        raise NotImplementedError("subclass must implement get_image()")

    def get_image_path(self):
        """Returns the image path for the current sample.

        Returns:
            the path to the image on disk
        """
        if not self.has_image_path:
            raise ValueError(
                "This '%s' does not provide image paths"
                % etau.get_class_name(self)
            )

        raise NotImplementedError("subclass must implement get_image_path()")

    def get_image_metadata(self):
        """Returns the image metadata for the current sample.

        Returns:
            a :class:`fiftyone.core.metadata.ImageMetadata` instance
        """
        if not self.has_image_metadata:
            raise ValueError(
                "This '%s' does not provide image metadata"
                % etau.get_class_name(self)
            )

        raise NotImplementedError(
            "subclass must implement get_image_metadata()"
        )

    def get_label(self):
        """Returns the label for the current sample.

        Returns:
            a :class:`fiftyone.core.labels.Label` instance, or a dictionary
            mapping field names to :class:`fiftyone.core.labels.Label`
            instances, or ``None`` if the sample is unlabeled
        """
        raise NotImplementedError("subclass must implement get_label()")


class LabeledVideoSampleParser(SampleParser):
    """Interface for :class:`SampleParser` instances that parse labeled video
    samples.

    The general recipe for using :class:`LabeledVideoSampleParser` instances
    is as follows::

        sample_parser = LabeledVideoSampleParser(...)

        for sample in samples:
            sample_parser.with_sample(sample)
            video_path = sample_parser.get_video_path()
            label = sample_parser.get_label()
            frames = sample_parser.get_frame_labels()

            if sample_parser.has_video_metadata:
                video_metadata = sample_parser.get_video_metadata()
    """

    @property
    def has_video_metadata(self):
        """Whether this parser produces
        :class:`fiftyone.core.metadata.VideoMetadata` instances for samples
        that it parses.
        """
        raise NotImplementedError("subclass must implement has_video_metadata")

    @property
    def label_cls(self):
        """The :class:`fiftyone.core.labels.Label` class(es) returned by this
        parser within the sample-level labels that it produces.

        This can be any of the following:

        -   a :class:`fiftyone.core.labels.Label` class. In this case, the
            parser is guaranteed to return sample-level labels of this type
        -   a list or tuple of :class:`fiftyone.core.labels.Label` classes. In
            this case, the parser can produce a single sample-level label field
            of any of these types
        -   a dict mapping keys to :class:`fiftyone.core.labels.Label` classes.
            In this case, the parser will return sample-level label
            dictionaries with keys and value-types specified by this
            dictionary. Not all keys need be present in the imported labels
        -   ``None``. In this case, the parser makes no guarantees about the
            sample-level labels that it may return
        """
        raise NotImplementedError("subclass must implement label_cls")

    @property
    def frame_labels_cls(self):
        """The :class:`fiftyone.core.labels.Label` class(es) returned by this
        parser within the frame labels that it produces.

        This can be any of the following:

        -   a :class:`fiftyone.core.labels.Label` class. In this case, the
            parser is guaranteed to return frame labels of this type
        -   a list or tuple of :class:`fiftyone.core.labels.Label` classes. In
            this case, the parser can produce a single frame label field of any
            of these types
        -   a dict mapping keys to :class:`fiftyone.core.labels.Label` classes.
            In this case, the parser will return frame label dictionaries with
            keys and value-types specified by this dictionary. Not all keys
            need be present in each frame
        -   ``None``. In this case, the parser makes no guarantees about the
            frame labels that it may return
        """
        raise NotImplementedError("subclass must implement frame_labels_cls")

    def get_video_path(self):
        """Returns the video path for the current sample.

        Returns:
            the path to the video on disk
        """
        raise NotImplementedError("subclass must implement get_video_path()")

    def get_video_metadata(self):
        """Returns the video metadata for the current sample.

        Returns:
            a :class:`fiftyone.core.metadata.ImageMetadata` instance
        """
        if not self.has_video_metadata:
            raise ValueError(
                "This '%s' does not provide video metadata"
                % etau.get_class_name(self)
            )

        raise NotImplementedError(
            "subclass must implement get_video_metadata()"
        )

    def get_label(self):
        """Returns the sample-level labels for the current sample.

        Returns:
            a :class:`fiftyone.core.labels.Label` instance, or a dictionary
            mapping field names to :class:`fiftyone.core.labels.Label`
            instances, or ``None`` if the sample has no sample-level labels
        """
        raise NotImplementedError("subclass must implement get_label()")

    def get_frame_labels(self):
        """Returns the frame labels for the current sample.

        Returns:
            a dictionary mapping frame numbers to dictionaries that map label
            fields to :class:`fiftyone.core.labels.Label` instances for each
            video frame, or ``None`` if the sample has no frame labels
        """
        raise NotImplementedError("subclass must implement get_frame_labels()")


class LabeledImageTupleSampleParser(LabeledImageSampleParser):
    """Generic sample parser that parses samples that are
    ``(image_or_path, label)`` tuples, where:

        - ``image_or_path`` is either an image that can be converted to numpy
          format via ``np.asarray()`` or the path to an image on disk

        - ``label`` is a :class:`fiftyone.core.labels.Label` instance

    This implementation provides a :meth:`_current_image` property that
    caches the image for the current sample, for efficiency in case multiple
    getters require access to the image (e.g., to normalize coordinates,
    compute metadata, etc).

    See the following subclasses of this parser for implementations that parse
    labels for common tasks:

        - Image classification: :class:`ImageClassificationSampleParser`
        - Object detection: :class:`ImageDetectionSampleParser`
        - Multitask image prediction: :class:`ImageLabelsSampleParser`
    """

    def __init__(self):
        super().__init__()
        self._current_image_cache = None

    @property
    def has_image_path(self):
        return True

    @property
    def has_image_metadata(self):
        return False

    @property
    def label_cls(self):
        return None

    def get_image(self):
        return self._current_image

    def get_image_path(self):
        image_or_path = self.current_sample[0]
        if etau.is_str(image_or_path):
            return image_or_path

        raise ValueError(
            "Cannot extract image path from samples that contain images"
        )

    def get_label(self):
        return self.current_sample[1]

    def clear_sample(self):
        super().clear_sample()
        self._current_image_cache = None

    @property
    def _current_image(self):
        if self._current_image_cache is None:
            self._current_image_cache = self._get_image()

        return self._current_image_cache

    def _get_image(self):
        image_or_path = self.current_sample[0]
        return self._parse_image(image_or_path)

    def _parse_image(self, image_or_path):
        if etau.is_str(image_or_path):
            return etai.read(image_or_path)

        return np.asarray(image_or_path)


class ImageClassificationSampleParser(LabeledImageTupleSampleParser):
    """Generic parser for image classification(s) samples whose labels are
    represented as :class:`fiftyone.core.labels.Classification` instances.

    This implementation supports samples that are ``(image_or_path, target)``
    tuples, where:

        -   ``image_or_path`` is either an image that can be converted to numpy
            format via ``np.asarray()`` or the path to an image on disk

        -   ``target`` can be any of the following:

            -   None, for unlabeled images
            -   a label string or list of label strings
            -   a class ID or list of class IDs, if ``classes`` is provided
            -   a dict or list of dicts of the following form::

                    {
                        "label": <label-or-target>,
                        "confidence": <confidence>,
                        "attributes": <optional-attributes>,
                    }

            -   a :class:`fiftyone.core.labels.Classification` or
                :class:`fiftyone.core.labels.Classifications` instance

    Args:
        classes (None): an optional list of class label strings. If provided,
            it is assumed that ``target`` contains class ID that should be
            mapped to label strings via ``classes[target]``
    """

    def __init__(self, classes=None):
        super().__init__()
        self.classes = classes

    @property
    def label_cls(self):
        return (fol.Classification, fol.Classifications)

    def get_label(self):
        """Returns the label for the current sample.

        Args:
            sample: the sample

        Returns:
            a :class:`fiftyone.core.labels.Classification` instance
        """
        target = self.current_sample[1]
        return self._parse_label(target)

    def _parse_label(self, target):
        if target is None:
            return None

        if isinstance(target, (fol.Classification, fol.Classifications)):
            return target

        is_list = isinstance(target, (list, tuple))

        if not is_list:
            target = [target]

        classifications = []
        for _target in target:
            if isinstance(_target, dict):
                label = _target.get("label", None)
                confidence = _target.get("confidence", None)
                attributes = _target.get("attributes", {})
            else:
                label = _target
                confidence = None
                attributes = {}

            try:
                label = self.classes[label]
            except:
                label = str(label)

            classification = fol.Classification(
                label=label, confidence=confidence, **attributes
            )
            classifications.append(classification)

        if is_list:
            return fol.Classifications(classifications=classifications)

        return classifications[0]


class ImageDetectionSampleParser(LabeledImageTupleSampleParser):
    """Generic parser for image detection samples whose labels are represented
    as :class:`fiftyone.core.labels.Detections` instances.

    This implementation supports samples that are
    ``(image_or_path, detections_or_path)`` tuples, where:

        - ``image_or_path`` is either an image that can be converted to numpy
          format via ``np.asarray()`` or the path to an image on disk

        - ``detections_or_path`` can be any of the following:

            -   None, for unlabeled images
            -   a list of detections in the following format::

                    [
                        {
                            "<label_field>": <label-or-target>,
                            "<bounding_box_field>": [
                                <top-left-x>, <top-left-y>, <width>, <height>
                            ],
                            "<confidence_field>": <optional-confidence>,
                            "<attributes_field>": {
                                <optional-name>: <optional-value>,
                                ...
                            }
                        },
                        ...
                    ]

                In the above, ``label-or-target`` is either a class ID
                (if ``classes`` is provided) or a label string, and the bounding
                box coordinates can either be relative coordinates in ``[0, 1]``
                (if ``normalized == True``) or absolute pixels coordinates
                (if ``normalized == False``). The confidence and attributes
                fields are optional for each sample.

                The input field names can be configured as necessary when
                instantiating the parser.

            -   the path on disk to a file in the above format
            -   a :class:`fiftyone.core.labels.Detections` instance

    Args:
        label_field ("label"): the name of the object label field in the
            target dicts
        bounding_box_field ("bounding_box"): the name of the bounding box field
            in the target dicts
        confidence_field (None): the name of the optional confidence field in
            the target dicts
        attributes_field (None): the name of the optional attributes field in
            the target dicts
        classes (None): an optional list of class label strings. If provided,
            it is assumed that the ``target`` values are class IDs that should
            be mapped to label strings via ``classes[target]``
        normalized (True): whether the bounding box coordinates are absolute
            pixel coordinates (``False``) or relative coordinates in [0, 1]
            (``True``)
    """

    def __init__(
        self,
        label_field="label",
        bounding_box_field="bounding_box",
        confidence_field=None,
        attributes_field=None,
        classes=None,
        normalized=True,
    ):
        super().__init__()
        self.label_field = label_field
        self.bounding_box_field = bounding_box_field
        self.confidence_field = confidence_field
        self.attributes_field = attributes_field
        self.classes = classes
        self.normalized = normalized

    @property
    def label_cls(self):
        return fol.Detections

    def get_label(self):
        """Returns the label for the current sample.

        Returns:
            a :class:`fiftyone.core.labels.Detections` instance
        """
        target = self.current_sample[1]

        if not self.normalized:
            # Absolute bounding box coordinates were provided, so we must have
            # the image to convert to relative coordinates
            img = self._current_image
        else:
            img = None

        return self._parse_label(target, img=img)

    def _parse_label(self, target, img=None):
        if target is None:
            return None

        if isinstance(target, fol.Detections):
            return target

        if etau.is_str(target):
            target = etas.read_json(target)

        return fol.Detections(
            detections=[self._parse_detection(obj, img=img) for obj in target]
        )

    def _parse_detection(self, obj, img=None):
        label = obj[self.label_field]

        try:
            label = self.classes[label]
        except:
            label = str(label)

        tlx, tly, w, h = self._parse_bbox(obj)

        if not self.normalized:
            height, width = img.shape[:2]
            tlx /= width
            tly /= height
            w /= width
            h /= height

        bounding_box = [tlx, tly, w, h]

        if self.confidence_field:
            confidence = obj.get(self.confidence_field, None)
        else:
            confidence = None

        if self.attributes_field:
            attributes = obj.get(self.attributes_field, {})
        else:
            attributes = {}

        return fol.Detection(
            label=label,
            bounding_box=bounding_box,
            confidence=confidence,
            **attributes,
        )

    def _parse_bbox(self, obj):
        return obj[self.bounding_box_field]


class ImageLabelsSampleParser(LabeledImageTupleSampleParser):
    """Generic parser for multitask image prediction samples whose labels are
    stored in ``eta.core.image.ImageLabels`` format.

    This implementation provided by this class supports samples that are
    ``(image_or_path, image_labels_or_path)`` tuples, where:

        - ``image_or_path`` is either an image that can be converted to numpy
          format via ``np.asarray()`` or the path to an image on disk

        - ``image_labels_or_path`` is an ``eta.core.image.ImageLabels``
          instance, an ``eta.core.frames.FrameLabels`` instance, a serialized
          dict representation of either, or the path to either on disk

    Args:
        prefix (None): a string prefix to prepend to each label name in the
            expanded label dictionary
        labels_dict (None): a dictionary mapping names of attributes/objects
            in the image labels to field names into which to expand them
        multilabel (False): whether to store attributes in a single
            :class:`fiftyone.core.labels.Classifications` instance
        skip_non_categorical (False): whether to skip non-categorical
            attributes (True) or cast them to strings (False)
    """

    def __init__(
        self,
        prefix=None,
        labels_dict=None,
        multilabel=False,
        skip_non_categorical=False,
    ):
        super().__init__()
        self.prefix = prefix
        self.labels_dict = labels_dict
        self.multilabel = multilabel
        self.skip_non_categorical = skip_non_categorical

    @property
    def label_cls(self):
        return None

    def get_label(self):
        """Returns the label for the current sample.

        Returns:
            a labels dictionary
        """
        labels = self.current_sample[1]
        return self._parse_label(labels)

    def _parse_label(self, labels):
        return foue.from_image_labels(
            labels,
            prefix=self.prefix,
            labels_dict=self.labels_dict,
            multilabel=self.multilabel,
            skip_non_categorical=self.skip_non_categorical,
        )


class FiftyOneImageClassificationSampleParser(ImageClassificationSampleParser):
    """Parser for samples in FiftyOne image classification datasets.

    See :ref:`this page <FiftyOneImageClassificationDataset-import>` for format
    details.

    Args:
        classes (None): an optional list of class label strings. If provided,
            it is assumed that ``target`` is a class ID that should be mapped
            to a label string via ``classes[target]``
    """

    def __init__(self, classes=None):
        super().__init__(classes=classes)


class FiftyOneTemporalDetectionSampleParser(LabeledVideoSampleParser):
    """Parser for samples in FiftyOne temporal detection datasets.

    See :ref:`this page <FiftyOneTemporalDetectionDataset-import>` for format
    details.

    Args:
        classes (None): an optional list of class label strings. If provided,
            it is assumed that ``target`` is a class ID that should be mapped
            to a label string via ``classes[target]``
        compute_metadata (False): whether to compute
            :class:`fiftyone.core.metadata.VideoMetadata` instances on-the-fly
            if :meth:`get_video_metadata` is called and no metadata is
            available
    """

    def __init__(self, classes=None, compute_metadata=False):
        super().__init__()
        self.classes = classes
        self.compute_metadata = compute_metadata
        self._current_metadata = None

    @property
    def has_video_metadata(self):
        return self.compute_metadata

    @property
    def label_cls(self):
        return fol.TemporalDetections

    @property
    def frame_labels_cls(self):
        return None

    def with_sample(self, sample, metadata=None):
        super().with_sample(sample)
        self._current_metadata = metadata

    def get_video_path(self):
        return self.current_sample[0]

    def get_video_metadata(self):
        if self._current_metadata is None and self.compute_metadata:
            video_path = self.current_sample[0]
            self._current_metadata = fom.VideoMetadata.build_for(video_path)

        return self._current_metadata

    def get_label(self):
        video_path, labels = self.current_sample

        if labels is None:
            return None

        if isinstance(labels, fol.TemporalDetections):
            return labels

        detections = []
        for label_dict in labels:
            label = label_dict["label"]

            try:
                label = self.classes[label]
            except:
                label = str(label)

            confidence = label_dict.get("confidence", None)
            attributes = label_dict.get("attributes", {})

            if "support" in label_dict:
                detection = fol.TemporalDetection(
                    label=label,
                    support=label_dict["support"],
                    confidence=confidence,
                    **attributes,
                )
            elif "timestamps" in label_dict:
                if self._current_metadata is not None:
                    metadata = self._current_metadata
                else:
                    metadata = fom.VideoMetadata.build_for(video_path)
                    self._current_metadata = metadata

                detection = fol.TemporalDetection.from_timestamps(
                    label_dict["timestamps"],
                    metadata=metadata,
                    label=label,
                    confidence=confidence,
                    **attributes,
                )
            else:
                raise ValueError(
                    "All temporal detection label dicts must have either "
                    "`support` or `timestamps` populated"
                )

            detections.append(detection)

        return fol.TemporalDetections(detections=detections)

    def get_frame_labels(self):
        return None

    def clear_sample(self):
        super().clear_sample()
        self._current_metadata = None


class FiftyOneImageDetectionSampleParser(ImageDetectionSampleParser):
    """Parser for samples in FiftyOne image detection datasets.

    See :ref:`this page <FiftyOneImageDetectionDataset-import>` for format
    details.

    Args:
        classes (None): an optional list of class label strings. If provided,
            it is assumed that the ``target`` values are class IDs that should
            be mapped to label strings via ``classes[target]``
    """

    def __init__(self, classes=None):
        super().__init__(
            label_field="label",
            bounding_box_field="bounding_box",
            confidence_field="confidence",
            attributes_field="attributes",
            classes=classes,
            normalized=True,
        )


class FiftyOneImageLabelsSampleParser(ImageLabelsSampleParser):
    """Parser for samples in FiftyOne image labels datasets.

    See :ref:`this page <FiftyOneImageLabelsDataset-import>` for format
    details.

    Args:
        prefix (None): a string prefix to prepend to each label name in the
            expanded label dictionary
        labels_dict (None): a dictionary mapping names of attributes/objects
            in the image labels to field names into which to expand them
        multilabel (False): whether to store attributes in a single
            :class:`fiftyone.core.labels.Classifications` instance
        skip_non_categorical (False): whether to skip non-categorical
            attributes (True) or cast them to strings (False)
    """

    pass


class VideoLabelsSampleParser(LabeledVideoSampleParser):
    """Generic parser for labeled video samples whose labels are represented in
    ``eta.core.video.VideoLabels`` format.

    This implementation provided by this class supports samples that are
    ``(video_path, video_labels_or_path)`` tuples, where:

        - ``video_path`` is the path to a video on disk

        - ``video_labels_or_path`` is an ``eta.core.video.VideoLabels``
          instance, a serialized dict representation of one, or the path to one
          on disk

    Args:
        prefix (None): a string prefix to prepend to each label name in the
            expanded sample/frame label dictionaries
        labels_dict (None): a dictionary mapping names of attributes/objects
            in the sample labels to field names into which to expand them. By
            default, all sample labels are loaded
        frame_labels_dict (None): a dictionary mapping names of
            attributes/objects in the frame labels to field names into which to
            expand them. By default, all frame labels are loaded
        multilabel (False): whether to store attributes in a single
            :class:`fiftyone.core.labels.Classifications` instance
        skip_non_categorical (False): whether to skip non-categorical
            attributes (True) or cast them to strings (False)
    """

    def __init__(
        self,
        prefix=None,
        labels_dict=None,
        frame_labels_dict=None,
        multilabel=False,
        skip_non_categorical=False,
    ):
        super().__init__()
        self.prefix = prefix
        self.labels_dict = labels_dict
        self.frame_labels_dict = frame_labels_dict
        self.multilabel = multilabel
        self.skip_non_categorical = skip_non_categorical

        self._curr_label = None
        self._curr_frames = None

    @property
    def has_video_metadata(self):
        return False

    @property
    def label_cls(self):
        return None

    @property
    def frame_labels_cls(self):
        return None

    def get_video_path(self):
        return self.current_sample[0]

    def get_label(self):
        self._parse_labels()
        return self._curr_label

    def get_frame_labels(self):
        self._parse_labels()
        return self._curr_frames

    def clear_sample(self):
        super().clear_sample()
        self._curr_label = None
        self._curr_frames = None

    def _parse_labels(self):
        if self._curr_label is not None or self._curr_frames is not None:
            return

        label, frames = foue.from_video_labels(
            self.current_sample[1],
            prefix=self.prefix,
            labels_dict=self.labels_dict,
            frame_labels_dict=self.frame_labels_dict,
            multilabel=self.multilabel,
            skip_non_categorical=self.skip_non_categorical,
        )

        self._curr_label = label
        self._curr_frames = frames


class FiftyOneVideoLabelsSampleParser(VideoLabelsSampleParser):
    """Parser for samples in FiftyOne video labels datasets.

    See :ref:`this page <FiftyOneVideoLabelsDataset-import>` for format
    details.

    Args:
        expand (True): whether to expand the labels for each frame into
            separate :class:`fiftyone.core.labels.Label` instances
        prefix (None): a string prefix to prepend to each label name in the
            expanded frame label dictionaries
        labels_dict (None): a dictionary mapping names of attributes/objects
            in the frame labels to field names into which to expand them
        multilabel (False): whether to store attributes in a single
            :class:`fiftyone.core.labels.Classifications` instance
        skip_non_categorical (False): whether to skip non-categorical
            attributes (True) or cast them to strings (False)
    """

    pass


class FiftyOneUnlabeledImageSampleParser(UnlabeledImageSampleParser):
    """Parser for :class:`fiftyone.core.sample.Sample` instances that contain
    images.

    Args:
        compute_metadata (False): whether to compute
            :class:`fiftyone.core.metadata.ImageMetadata` instances on-the-fly
            if :meth:`get_image_metadata` is called and no metadata is
            available
    """

    def __init__(self, compute_metadata=False):
        super().__init__()
        self.compute_metadata = compute_metadata

    @property
    def has_image_path(self):
        return True

    @property
    def has_image_metadata(self):
        return True

    def get_image(self):
        fov.validate_image_sample(self.current_sample)
        return etai.read(self.current_sample.filepath)

    def get_image_path(self):
        fov.validate_image_sample(self.current_sample)
        return self.current_sample.filepath

    def get_image_metadata(self):
        fov.validate_image_sample(self.current_sample)
        metadata = self.current_sample.metadata
        if metadata is None and self.compute_metadata:
            metadata = fom.ImageMetadata.build_for(
                self.current_sample.filepath
            )

        return metadata


class FiftyOneLabeledImageSampleParser(LabeledImageSampleParser):
    """Parser for :class:`fiftyone.core.sample.Sample` instances that contain
    labeled images.

    Args:
        label_field: the name of the label field to parse, or a dictionary
            mapping label field names to keys for the return label dictionaries
        label_fcn (None): an optional function or dictionary mapping label
            field names to functions (must match ``label_field``) to apply to
            each label before returning it
        compute_metadata (False): whether to compute
            :class:`fiftyone.core.metadata.ImageMetadata` instances on-the-fly
            if :meth:`get_image_metadata` is called and no metadata is
            available
    """

    def __init__(self, label_field, label_fcn=None, compute_metadata=False):
        super().__init__()
        self.label_field = label_field
        self.label_fcn = label_fcn
        self.compute_metadata = compute_metadata

    @property
    def has_image_path(self):
        return True

    @property
    def has_image_metadata(self):
        return True

    @property
    def label_cls(self):
        return None

    def get_image(self):
        fov.validate_image_sample(self.current_sample)
        return etai.read(self.current_sample.filepath)

    def get_image_path(self):
        fov.validate_image_sample(self.current_sample)
        return self.current_sample.filepath

    def get_image_metadata(self):
        fov.validate_image_sample(self.current_sample)
        metadata = self.current_sample.metadata
        if metadata is None and self.compute_metadata:
            metadata = fom.ImageMetadata.build_for(
                self.current_sample.filepath
            )

        return metadata

    def get_label(self):
        sample = self.current_sample
        label_field = self.label_field
        label_fcn = self.label_fcn

        if isinstance(label_field, dict):
            if label_fcn is not None:
                label = {}
                for k, v in label_field.items():
                    f = label_fcn.get(k, None)
                    if f is not None:
                        label[v] = f(sample[k])
                    else:
                        label[v] = sample[k]
            else:
                label = {v: sample[k] for k, v in label_field.items()}
        else:
            label = sample[label_field]
            if label_fcn is not None:
                label = label_fcn(label)

        return label


class ExtractClipsMixin(object):
    """Mixin for sample parsers that extract clips from
    :class:`fiftyone.core.clips.ClipView` instances.

    Args:
        compute_metadata (False): whether to compute
            :class:`fiftyone.core.metadata.VideoMetadata` instances on-the-fly
            when no pre-computed metadata is available
        export_media (True): whether to actually write clips when their paths
            are requested
        clip_dir (None): a directory to write clips. Only applicable when
            parsing :class:`fiftyone.core.clips.ClipView` instances
        video_format (None): the video format to use when writing video clips
            to disk. By default, ``fiftyone.config.default_video_ext`` is used
    """

    def __init__(
        self,
        compute_metadata=False,
        export_media=True,
        clip_dir=None,
        video_format=None,
    ):
        if video_format is None:
            video_format = fo.config.default_video_ext

        self.compute_metadata = compute_metadata
        self.export_media = export_media
        self.clip_dir = clip_dir
        self.video_format = video_format

        self._curr_clip_path = None

    def _get_clip_path(self, sample):
        video_path = sample.filepath
        basename, ext = os.path.splitext(os.path.basename(video_path))

        if self.export_media:
            if self.clip_dir is None:
                self.clip_dir = etau.make_temp_dir()

            dirname = self.clip_dir
            ext = self.video_format
        else:
            dirname = os.path.dirname(video_path)

        clip_name = "%s-clip-%d-%d%s" % (
            basename,
            sample.support[0],
            sample.support[1],
            ext,
        )
        clip_path = os.path.join(dirname, clip_name)

        if self.export_media:
            self._curr_clip_path = clip_path
            fouv.extract_clip(
                video_path,
                clip_path,
                support=sample.support,
                metadata=sample.metadata,
            )
        else:
            self._curr_clip_path = None

        return clip_path

    def _get_clip_metadata(self, sample):
        if not self.compute_metadata or self._curr_clip_path is None:
            return None

        return fom.VideoMetadata.build_for(self._curr_clip_path)


class FiftyOneUnlabeledVideoSampleParser(
    ExtractClipsMixin, UnlabeledVideoSampleParser
):
    """Parser for :class:`fiftyone.core.sample.Sample` instances that contain
    videos.

    This class also supports :class:`fiftyone.core.clips.ClipView` instances.

    Args:
        compute_metadata (False): whether to compute
            :class:`fiftyone.core.metadata.VideoMetadata` instances on-the-fly
            if :meth:`get_video_metadata` is called and no metadata is
            available
        export_media (True): whether to write clips when :meth:`get_video_path`
            is called
        clip_dir (None): a directory to write clips. Only applicable when
            parsing :class:`fiftyone.core.clips.ClipView` instances
        video_format (None): the video format to use when writing video clips
            to disk. By default, ``fiftyone.config.default_video_ext`` is used
    """

    def __init__(
        self,
        compute_metadata=False,
        export_media=True,
        clip_dir=None,
        video_format=None,
    ):
        ExtractClipsMixin.__init__(
            self,
            compute_metadata=compute_metadata,
            export_media=export_media,
            clip_dir=clip_dir,
            video_format=video_format,
        )
        UnlabeledVideoSampleParser.__init__(self)

    @property
    def has_video_metadata(self):
        return True

    def get_video_path(self):
        if isinstance(self.current_sample, foc.ClipView):
            return self._get_clip_path(self.current_sample)

        return self.current_sample.filepath

    def get_video_metadata(self):
        if isinstance(self.current_sample, foc.ClipView):
            return self._get_clip_metadata(self.current_sample)

        metadata = self.current_sample.metadata
        if metadata is None and self.compute_metadata:
            metadata = fom.VideoMetadata.build_for(
                self.current_sample.filepath
            )

        return metadata


class FiftyOneLabeledVideoSampleParser(
    ExtractClipsMixin, LabeledVideoSampleParser
):
    """Parser for :class:`fiftyone.core.sample.Sample` instances that contain
    labeled videos.

    This class also supports :class:`fiftyone.core.clips.ClipView` instances.

    Args:
        label_field (None): the name of a label field to parse, or a dictionary
            mapping label field names to output keys to use in the returned
            sample-level labels dictionary
        frame_labels_field (None): the name of a frame label field to parse, or
            a dictionary mapping field names to output keys describing the
            frame label fields to export
        label_fcn (None): an optional function or dictionary mapping label
            field names to functions (must match ``label_field``) to apply to
            each sample label before returning it
        frame_labels_fcn (None): an optional function or dictionary mapping
            frame label field names to functions (must match
            ``frame_labels_field``) to apply to each frame label before
            returning it
        compute_metadata (False): whether to compute
            :class:`fiftyone.core.metadata.VideoMetadata` instances on-the-fly
            if :meth:`get_video_metadata` is called and no metadata is
            available
        export_media (True): whether to write clips when :meth:`get_video_path`
            is called
        clip_dir (None): a directory to write clips. Only applicable when
            parsing :class:`fiftyone.core.clips.ClipView` instances
        video_format (None): the video format to use when writing video clips
            to disk. By default, ``fiftyone.config.default_video_ext`` is used
    """

    def __init__(
        self,
        label_field=None,
        frame_labels_field=None,
        label_fcn=None,
        frame_labels_fcn=None,
        compute_metadata=False,
        export_media=True,
        clip_dir=None,
        video_format=None,
    ):
        frame_labels_dict, frame_fcn_dict = self._parse_frame_args(
            frame_labels_field, frame_labels_fcn
        )

        ExtractClipsMixin.__init__(
            self,
            compute_metadata=compute_metadata,
            export_media=export_media,
            clip_dir=clip_dir,
            video_format=video_format,
        )
        LabeledVideoSampleParser.__init__(self)

        self.label_field = label_field
        self.frame_labels_dict = frame_labels_dict
        self.label_fcn = label_fcn
        self.frame_fcn_dict = frame_fcn_dict

    @property
    def has_video_metadata(self):
        return True

    @property
    def label_cls(self):
        return None

    @property
    def frame_labels_cls(self):
        return None

    def get_video_path(self):
        if isinstance(self.current_sample, foc.ClipView):
            return self._get_clip_path(self.current_sample)

        return self.current_sample.filepath

    def get_video_metadata(self):
        if isinstance(self.current_sample, foc.ClipView):
            return self._get_clip_metadata(self.current_sample)

        metadata = self.current_sample.metadata
        if metadata is None and self.compute_metadata:
            metadata = fom.VideoMetadata.build_for(
                self.current_sample.filepath
            )

        return metadata

    def get_label(self):
        sample = self.current_sample
        label_field = self.label_field
        label_fcn = self.label_fcn

        if label_field is None:
            return None

        if isinstance(label_field, dict):
            if label_fcn is not None:
                label = {}
                for k, v in label_field.items():
                    f = label_fcn.get(k, None)
                    if f is not None:
                        label[v] = f(sample[k])
                    else:
                        label[v] = sample[k]
            else:
                label = {v: sample[k] for k, v in label_field.items()}
        else:
            label = sample[label_field]
            if label_fcn is not None:
                label = label_fcn(label)

        return label

    def get_frame_labels(self):
        if isinstance(self.current_sample, foc.ClipView):
            df = self.current_sample.support[0] - 1
        else:
            df = 0

        frames = self.current_sample.frames
        frame_labels_dict = self.frame_labels_dict
        frame_fcn_dict = self.frame_fcn_dict

        if frame_labels_dict is None:
            return None

        if frame_fcn_dict is not None:
            new_frames = {}
            for frame_number, frame in frames.items():
                new_frame = {}
                for k, v in frame_labels_dict.items():
                    f = frame_fcn_dict.get(k, None)
                    if f is not None:
                        new_frame[v] = f(frame[k])
                    else:
                        new_frame[v] = frame[k]

                new_frames[frame_number - df] = new_frame
        else:
            new_frames = {}
            for frame_number, frame in frames.items():
                new_frames[frame_number - df] = {
                    v: frame[k] for k, v in frame_labels_dict.items()
                }

        return new_frames

    @staticmethod
    def _parse_frame_args(frame_labels_field, frame_labels_fcn):
        if frame_labels_field is None:
            return None, None

        if not isinstance(frame_labels_field, dict):
            label_field = frame_labels_field
            frame_labels_dict = {label_field: label_field}
            if frame_labels_fcn is not None:
                frame_fcn_dict = {label_field: frame_labels_fcn}
            else:
                frame_fcn_dict = None
        else:
            frame_labels_dict = frame_labels_field
            frame_fcn_dict = frame_labels_fcn

        return frame_labels_dict, frame_fcn_dict



"""
FiftyOne Server events

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import defaultdict
from dataclasses import asdict, dataclass
import typing as t
from datetime import datetime

import asyncio
from sse_starlette import ServerSentEvent
from starlette.requests import Request

import fiftyone as fo
import fiftyone.core.context as focx
from fiftyone.core.json import FiftyOneJSONEncoder
from fiftyone.core.session.events import (
    add_screenshot,
    CaptureNotebookCell,
    CloseSession,
    DeactivateNotebookCell,
    ReactivateNotebookCell,
    dict_factory,
    EventType,
    ListenPayload,
    StateUpdate,
)
import fiftyone.core.state as fos
from fiftyone.server.query import serialize_dataset


@dataclass(frozen=True)
class Listener:
    queue: asyncio.Queue
    subscription: str


_listeners: t.Dict[str, t.Set[Listener]] = defaultdict(set)
_requests: t.Dict[str, t.Set[t.Tuple[str, Listener]]] = {}
_polling_listener: t.Optional[
    t.Tuple[str, t.Set[t.Tuple[str, Listener]]]
] = None
_state: t.Optional[fos.StateDescription] = None
_app_count = 0


async def dispatch_event(
    subscription: t.Optional[str], event: EventType
) -> None:
    """Dispatch an event to all listeners registed for the server process

    Args:
        subscription: the calling subscription id
        event: the event
    """
    if isinstance(event, CaptureNotebookCell) and focx.is_databricks_context():
        add_screenshot(event.subscription, event.src)
        return

    if isinstance(event, StateUpdate):
        global _state
        _state = event.state

    if isinstance(event, ReactivateNotebookCell):
        await dispatch_event(subscription, DeactivateNotebookCell())

    for listener in _listeners[event.get_event_name()]:
        if listener.subscription == subscription:
            continue

        listener.queue.put_nowait((datetime.now(), event))


async def add_event_listener(
    request: Request, payload: ListenPayload
) -> t.AsyncIterator:
    """Add an event listener to the server

    Args:
        request: the event source request
        payload: the initialization payload

    Returns:
        A server sent event source
    """
    data = await _initialize_listener(payload)
    try:
        if data.is_app:
            d = asdict(
                StateUpdate(state=data.state),
                dict_factory=dict_factory,
            )
            if data.state.dataset is not None:
                d["dataset"] = await serialize_dataset(
                    data.state.dataset.name,
                    data.state.view._serialize()
                    if data.state.view is not None
                    else [],
                )

            yield ServerSentEvent(
                event=StateUpdate.get_event_name(),
                data=FiftyOneJSONEncoder.dumps(d),
            )

        while True:
            disconnected = await request.is_disconnected()
            if disconnected:
                await _disconnect(
                    data.is_app,
                    data.request_listeners,
                )
                break

            events: t.List[t.Tuple[datetime, EventType]] = []
            for _, listener in data.request_listeners:
                if listener.queue.qsize():
                    events.append(listener.queue.get_nowait())

            events = sorted(events, key=lambda event: event[0])

            for (_, event) in events:
                d = asdict(event, dict_factory=dict_factory)

                if (
                    data.is_app
                    and isinstance(event, StateUpdate)
                    and event.state.dataset is not None
                ):
                    d["dataset"] = await serialize_dataset(
                        event.state.dataset.name,
                        event.state.view._serialize()
                        if event.state.view is not None
                        else [],
                    )

                yield ServerSentEvent(
                    event=event.get_event_name(),
                    data=FiftyOneJSONEncoder.dumps(d),
                )

            await asyncio.sleep(0.2)

    except asyncio.CancelledError as e:
        await _disconnect(data.is_app, data.request_listeners)
        raise e


async def dispatch_polling_event_listener(
    request: Request, payload: ListenPayload
) -> t.Dict:
    """Polling event listener interface

    Note:

        The polling event listener is a singleton, and is only a fallback for
        Google's Colaboratory connections

    Args:
        request: the event source request
        payload: the initialization payload

    Returns:
        A server sent event source
    """
    global _polling_listener
    global _requests
    sub = None

    if _polling_listener is not None:
        sub, _ = _polling_listener

    if sub != payload.subscription and payload.subscription not in _requests:
        data = await _initialize_listener(payload)
        _polling_listener = (payload.subscription, data.request_listeners)

        return {
            "events": [
                {
                    "event": StateUpdate.get_event_name(),
                    "data": asdict(
                        StateUpdate(state=data.state),
                        dict_factory=dict_factory,
                    ),
                }
            ]
        }

    events: t.List[t.Tuple[datetime, EventType]] = []
    disconnect = False
    for _, listener in _requests[payload.subscription]:
        while listener.queue.qsize():
            event_tuple = listener.queue.get_nowait()
            if isinstance(event_tuple[1], DeactivateNotebookCell):
                disconnect = True

            events.append(event_tuple)

    if disconnect:
        del _requests[payload.subscription]

    events = sorted(events, key=lambda event: event[0])
    return {
        "events": [
            {
                "event": e.get_event_name(),
                "data": asdict(e, dict_factory=dict_factory),
            }
            for (_, e) in events
        ],
    }


def get_state() -> fos.StateDescription:
    """Get the current state description singleton on the server

    Returns:
        the :class:`fiftyone.core.state.StateDescription` server singleton
    """
    global _state
    if _state is None:
        _state = fos.StateDescription()

    return _state


async def _disconnect(
    is_app: bool, listeners: t.Set[t.Tuple[str, Listener]]
) -> None:
    for event_name, listener in listeners:
        _listeners[event_name].remove(listener)

    if is_app:
        global _app_count
        _app_count -= 1

        if not _app_count and focx._get_context() == focx._NONE:
            await dispatch_event(None, CloseSession())


@dataclass
class InitializedListener:
    is_app: bool
    request_listeners: t.Set[t.Tuple[str, Listener]]
    state: fos.StateDescription


async def _initialize_listener(payload: ListenPayload) -> InitializedListener:
    state = get_state()
    is_app = not isinstance(payload.initializer, fos.StateDescription)
    if is_app:
        global _app_count
        _app_count += 1

    current = state.dataset.name if state.dataset is not None else None
    if is_app and payload.initializer != current:
        if payload.initializer is not None:
            try:
                state.dataset = fo.load_dataset(payload.initializer)
            except:
                state.dataset = None
            state.selected = []
            state.selected_labels = []
            state.view = None

            await dispatch_event(payload.subscription, StateUpdate(state))

    elif not is_app:
        state = payload.initializer
        await dispatch_event(payload.subscription, StateUpdate(state))

    request_listeners: t.Set[t.Tuple[str, Listener]] = set()
    for event_name in payload.events:
        listener = Listener(
            queue=asyncio.LifoQueue(maxsize=1000),
            subscription=payload.subscription,
        )
        _listeners[event_name].add(listener)
        request_listeners.add((event_name, listener))

    global _requests
    _requests[payload.subscription] = request_listeners

    return InitializedListener(is_app, request_listeners, state)



"""
FiftyOne service utilities.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""

import psutil

from fiftyone.service.ipc import send_request


def describe_process(process):
    """Returns a detailed description of a process.

    Args:
        process (psutil.Process)

    Returns:
        str
    """
    try:
        details = repr(process.cmdline())
    except psutil.Error:
        try:
            details = process.name()
        except psutil.Error:
            details = "unknown"
    return "Process %i (%s)" % (process.pid, details)


def _is_wrapper_process(process):
    """Returns true if the specified process is a wrapper around a single
    child process with the same arguments.

    This can happen on Windows when a Python subprocess is created. These
    processes should generally be ignored.

    Args:
        process (psutil.Process)

    Returns:
        bool
    """
    try:
        children = process.children()
        if len(children) != 1:
            return False
        if process.cmdline()[1:] == children[0].cmdline()[1:]:
            return True
    except psutil.Error:
        pass
    return False


def normalize_wrapper_process(process):
    """Returns the given process, or its child if it is a wrapper processes.

    See _is_wrapper_process() for details.

    Args:
        process (psutil.Process)

    Returns:
        psutil.Process
    """
    if _is_wrapper_process(process):
        try:
            return process.children()[0]
        except IndexError:
            pass
    return process


def find_processes_by_args(args):
    """Finds a process with the specified command-line arguments.

    Only processes for the current user will be returned.

    Args:
        args (list[str]): a list of arguments, in the order to search for

    Returns:
        generator of psutil.Process objects
    """
    if not isinstance(args, list):
        raise TypeError("args must be list")
    if not args:
        raise ValueError("empty search")

    current_username = psutil.Process().username()
    for p in psutil.process_iter(["cmdline", "username"]):
        try:
            if p.info["username"] == current_username and p.info["cmdline"]:
                cmdline = p.info["cmdline"]
                for i in range(len(cmdline) - len(args) + 1):
                    if cmdline[i : i + len(args)] == args:
                        if not _is_wrapper_process(p):
                            yield p
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            pass


def get_listening_tcp_ports(process):
    """Retrieves a list of TCP ports that the specified process is listening on.

    Args:
        process (psutil.Process): the process to check

    Returns:
        generator of integers
    """
    for conn in process.connections(kind="tcp"):
        if (
            not conn.raddr  # not connected to a remote socket
            and conn.status == psutil.CONN_LISTEN
        ):
            yield conn.laddr[1]  # port


def send_ipc_message(process, message):
    """Sends a message to a process's IPCServer.

    Args:
        process (psutil.Process): process to send the message to
        message (any type)

    Returns:
        response (any type)
    """
    try:
        port = next(get_listening_tcp_ports(process))
    except StopIteration:
        raise IOError("Process %i has no listening server" % process.pid)
    return send_request(port, message)



"""
FiftyOne Server JIT metadata utilities

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from enum import Enum
import logging
import shutil
import struct
import typing as t

import asyncio
import aiofiles
import strawberry as gql

import eta.core.serial as etas
import eta.core.utils as etau
import eta.core.video as etav
from fiftyone.core.collections import SampleCollection

import fiftyone.core.media as fom

logger = logging.getLogger(__name__)

_FFPROBE_BINARY_PATH = shutil.which("ffprobe")


@gql.enum
class MediaType(Enum):
    image = "image"
    group = "group"
    point_cloud = "point-cloud"
    video = "video"


async def get_metadata(
    collection: SampleCollection,
    sample: t.Dict,
    media_type: t.Dict[str, t.Dict],
    metadata_cache: t.Dict[str, t.Dict[str, str]],
    url_cache: t.Dict[str, str],
):
    """Gets the metadata for the given local media file.

    Args:
        filepath: the path to the file
        media_type: the file's media type
        metadata (None): a pre-existing metadata dict to use if possible

    Returns:
        metadata dict
    """
    filepath = sample["filepath"]
    is_video = media_type == fom.VIDEO
    metadata = sample.get("metadata", None)
    urls = _create_media_urls(collection, sample, url_cache)

    # If sufficient pre-existing metadata exists, use it
    if filepath not in metadata_cache and metadata:
        if is_video:
            width = metadata.get("frame_width", None)
            height = metadata.get("frame_height", None)
            frame_rate = metadata.get("frame_rate", None)

            if width and height and frame_rate:
                metadata_cache[sample["filepath"]] = dict(
                    aspect_ratio=width / height,
                    frame_rate=frame_rate,
                )

        else:
            width = metadata.get("width", None)
            height = metadata.get("height", None)

            if width and height:
                metadata_cache[sample["filepath"]] = dict(
                    aspect_ratio=width / height,
                )

    if filepath not in metadata_cache:
        try:
            # Retrieve media metadata from disk
            metadata_cache[filepath] = await read_metadata(filepath, is_video)
        except Exception as exc:
            # Immediately fail so the user knows they should install FFmpeg
            if isinstance(exc, FFmpegNotFoundException):
                raise exc

            # Something went wrong (ie non-existent file), so we gracefully return
            # some placeholder metadata so the App grid can be rendered
            if is_video:
                metadata_cache[filepath] = dict(aspect_ratio=1, frame_rate=30)
            else:
                metadata_cache[filepath] = dict(aspect_ratio=1)

    return dict(urls=urls, **metadata_cache[filepath])


async def read_metadata(filepath, is_video):
    """Calculates the metadata for the given local media path.

    Args:
        filepath: a filepath
        is_video: whether the file is a video

    Returns:
        dict
    """
    if is_video:
        info = await get_stream_info(filepath)
        return dict(
            aspect_ratio=info.frame_size[0] / info.frame_size[1],
            frame_rate=info.frame_rate,
        )

    async with aiofiles.open(filepath, "rb") as f:
        width, height = await get_image_dimensions(f)
        return dict(aspect_ratio=width / height)


class Reader(object):
    """Asynchronous file-like reader.

    Args:
        content: a :class:`aiohttp.StreamReader`
    """

    def __init__(self, content):
        self._data = b""
        self._content = content

    async def read(self, bytes):
        data = await self._content.read(bytes)
        self._data += data
        return data

    async def seek(self, bytes):
        delta = bytes - len(self._data)
        if delta < 0:
            data = self._data[delta:]
            self._data = data[:delta]
            self._content.unread_data(data)
        else:
            self._data += await self._content.read(delta)


async def get_stream_info(path):
    """Returns a :class:`eta.core.video.VideoStreamInfo` instance for the
    provided video path.

    Args:
        path: a video filepath

    Returns:
        a :class:`eta.core.video.VideoStreamInfo`
    """
    if _FFPROBE_BINARY_PATH is None:
        raise FFmpegNotFoundException(
            "You must have ffmpeg installed on your machine in order to view "
            "video datasets in the App, but we failed to find it"
        )

    proc = await asyncio.create_subprocess_exec(
        _FFPROBE_BINARY_PATH,
        "-loglevel",
        "error",
        "-show_format",
        "-show_streams",
        "-print_format",
        "json",
        "-i",
        path,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    stdout, stderr = await proc.communicate()
    if stderr:
        raise ValueError(stderr)

    info = etas.load_json(stdout.decode("utf8"))

    video_streams = [s for s in info["streams"] if s["codec_type"] == "video"]
    num_video_streams = len(video_streams)
    if num_video_streams == 1:
        stream_info = video_streams[0]
    elif num_video_streams == 0:
        logger.debug("No video stream found; defaulting to first stream")
        stream_info = info["streams"][0]
    else:
        logger.debug("Found multiple video streams; using first stream")
        stream_info = video_streams[0]

    format_info = info["format"]
    mime_type = etau.guess_mime_type(path)

    return etav.VideoStreamInfo(stream_info, format_info, mime_type=mime_type)


async def get_image_dimensions(input):
    """Gets the dimensions of an image from its file-like asynchronous byte
    stream.

    Args:
        input: file-like object with async read and seek methods

    Returns:
        the ``(width, height)``
    """
    height = -1
    width = -1
    data = await input.read(26)
    size = len(data)

    if (size >= 10) and data[:6] in (b"GIF87a", b"GIF89a"):
        # GIFs
        w, h = struct.unpack("<HH", data[6:10])
        width = int(w)
        height = int(h)
    elif (
        (size >= 24)
        and data.startswith(b"\211PNG\r\n\032\n")
        and (data[12:16] == b"IHDR")
    ):
        # PNGs
        w, h = struct.unpack(">LL", data[16:24])
        width = int(w)
        height = int(h)
    elif (size >= 16) and data.startswith(b"\211PNG\r\n\032\n"):
        # older PNGs
        w, h = struct.unpack(">LL", data[8:16])
        width = int(w)
        height = int(h)
    elif (size >= 2) and data.startswith(b"\377\330"):
        await input.seek(2)
        b = await input.read(1)
        while b and ord(b) != 0xDA:
            while ord(b) != 0xFF:
                b = await input.read(1)
            while ord(b) == 0xFF:
                b = await input.read(1)
            if ord(b) >= 0xC0 and ord(b) <= 0xC3:
                await input.read(3)
                tmp = await input.read(4)
                h, w = struct.unpack(">HH", tmp)
                break
            else:
                tmp = await input.read(2)
                await input.read(int(struct.unpack(">H", tmp)[0]) - 2)
            b = await input.read(1)
        width = int(w)
        height = int(h)
    elif (size >= 26) and data.startswith(b"BM"):
        # BMP
        headersize = struct.unpack("<I", data[14:18])[0]
        if headersize == 12:
            w, h = struct.unpack("<HH", data[18:22])
            width = int(w)
            height = int(h)
        elif headersize >= 40:
            w, h = struct.unpack("<ii", data[18:26])
            width = int(w)
            # as h is negative when stored upside down
            height = abs(int(h))
        else:
            raise MetadataException(
                "Unkown DIB header size: %s" % str(headersize)
            )
    elif (size >= 8) and data[:4] in (b"II\052\000", b"MM\000\052"):
        # Standard TIFF, big- or little-endian
        # BigTIFF and other different but TIFF-like formats are not
        # supported currently
        byteOrder = data[:2]
        boChar = ">" if byteOrder == "MM" else "<"
        # maps TIFF type id to size (in bytes)
        # and python format char for struct
        tiffTypes = {
            1: (1, boChar + "B"),  # BYTE
            2: (1, boChar + "c"),  # ASCII
            3: (2, boChar + "H"),  # SHORT
            4: (4, boChar + "L"),  # LONG
            5: (8, boChar + "LL"),  # RATIONAL
            6: (1, boChar + "b"),  # SBYTE
            7: (1, boChar + "c"),  # UNDEFINED
            8: (2, boChar + "h"),  # SSHORT
            9: (4, boChar + "l"),  # SLONG
            10: (8, boChar + "ll"),  # SRATIONAL
            11: (4, boChar + "f"),  # FLOAT
            12: (8, boChar + "d"),  # DOUBLE
        }
        ifdOffset = struct.unpack(boChar + "L", data[4:8])[0]

        countSize = 2
        await input.seek(ifdOffset)
        ec = await input.read(countSize)
        ifdEntryCount = struct.unpack(boChar + "H", ec)[0]
        # 2 bytes: TagId + 2 bytes: type + 4 bytes: count of values + 4
        # bytes: value offset
        ifdEntrySize = 12
        for i in range(ifdEntryCount):
            entryOffset = ifdOffset + countSize + i * ifdEntrySize
            await input.seek(entryOffset)
            tag = await input.read(2)
            tag = struct.unpack(boChar + "H", tag)[0]
            if tag == 256 or tag == 257:
                # if type indicates that value fits into 4 bytes, value
                # offset is not an offset but value itself
                type = await input.read(2)
                type = struct.unpack(boChar + "H", type)[0]
                if type not in tiffTypes:
                    raise MetadataException("Unable to read metadata")
                typeSize = tiffTypes[type][0]
                typeChar = tiffTypes[type][1]
                await input.seek(entryOffset + 8)
                value = await input.read(typeSize)
                value = int(struct.unpack(typeChar, value)[0])
                if tag == 256:
                    width = value
                else:
                    height = value
            if width > -1 and height > -1:
                break

    elif size >= 2:
        await input.seek(0)
        reserved = await input.read(2)
        if 0 != struct.unpack("<H", reserved)[0]:
            raise MetadataException("Unable to read metadata")
        format = await input.read(2)
        if 1 != struct.unpack("<H", format)[0]:
            raise MetadataException("Unable to read metadata")
        num = await input.read(2)
        num = struct.unpack("<H", num)[0]

        # http://msdn.microsoft.com/en-us/library/ms997538.aspx
        w = await input.read(1)
        h = await input.read(1)
        width = ord(w)
        height = ord(h)

    return width, height


class MetadataException(Exception):
    """Exception raised when metadata for a media file cannot be computed."""

    pass


class FFmpegNotFoundException(RuntimeError):
    """Exception raised when FFmpeg or FFprobe cannot be found."""

    pass


def _create_media_urls(
    collection: SampleCollection, sample: t.Dict, cache: t.Dict
) -> t.Dict[str, str]:
    media_fields = collection.app_config.media_fields
    media_urls = []

    for field in media_fields:
        path = sample.get(field, None)
        if path not in cache:
            cache[path] = path

        media_urls.append(dict(field=field, url=path))

    return media_urls



"""
FiftyOne v0.16.0 revision.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""


def up(db, dataset_name):
    match_d = {"name": dataset_name}
    dataset_dict = db.datasets.find_one(match_d)

    media_type = dataset_dict.get("media_type", None)

    for field in dataset_dict.get("sample_fields", []):
        name = field.get("name", None)
        ftype = field.get("ftype", None)
        embedded_doc_type = field.get("embedded_doc_type", None)

        # Upgrade `metadata` field to `media_type`-aware subclass
        if name == "metadata":
            if media_type == "image":
                embedded_doc_type = "fiftyone.core.metadata.ImageMetadata"
            elif media_type == "video":
                embedded_doc_type = "fiftyone.core.metadata.VideoMetadata"
            else:
                embedded_doc_type = "fiftyone.core.metadata.Metadata"

            field["embedded_doc_type"] = embedded_doc_type

        # Populate embedded field schemas
        if ftype == "fiftyone.core.fields.EmbeddedDocumentField":
            try:
                coll = db[dataset_dict["sample_collection_name"]]
                field["fields"] = _infer_fields(coll, name, embedded_doc_type)
            except Exception as e:
                print(
                    "Failed to infer schema of sample field '%s' of type "
                    "'%s': %s" % (name, embedded_doc_type, e)
                )
                field["fields"] = []
        else:
            field["fields"] = []

    for field in dataset_dict.get("frame_fields", []):
        name = field.get("name", None)
        ftype = field.get("ftype", None)
        embedded_doc_type = field.get("embedded_doc_type", None)

        # Populate embedded field schemas
        if ftype == "fiftyone.core.fields.EmbeddedDocumentField":
            try:
                coll = db[dataset_dict["frame_collection_name"]]
                field["fields"] = _infer_fields(coll, name, embedded_doc_type)
            except Exception as e:
                print(
                    "Failed to infer schema of frame field '%s' of type "
                    "'%s': %s" % (name, embedded_doc_type, e)
                )
                field["fields"] = []
        else:
            field["fields"] = []

    dataset_dict["app_sidebar_groups"] = None

    db.datasets.replace_one(match_d, dataset_dict)


def down(db, dataset_name):
    match_d = {"name": dataset_name}
    dataset_dict = db.datasets.find_one(match_d)

    for field in dataset_dict.get("sample_fields", []):
        if field.get("name", None) == "metadata":
            field["embedded_doc_type"] = "fiftyone.core.metadata.Metadata"

        field.pop("fields", None)

    for field in dataset_dict.get("frame_fields", []):
        field.pop("fields", None)

    dataset_dict.pop("app_sidebar_groups", None)

    db.datasets.replace_one(match_d, dataset_dict)


def _infer_fields(coll, name, embedded_doc_type):
    fields = _do_infer_fields(coll, name, embedded_doc_type)

    list_field = _LABEL_LIST_FIELDS.get(embedded_doc_type, None)

    for field in fields:
        if (
            list_field is not None
            and field.get("name", None) == list_field
            and field.get("ftype", None) == "fiftyone.core.fields.ListField"
        ):
            list_path = name + "." + list_field
            ltype = embedded_doc_type[:-1]  # remove "s"
            field["subfield"] = "fiftyone.core.fields.EmbeddedDocumentField"
            field["embedded_doc_type"] = ltype
            field["fields"] = _do_infer_fields(coll, list_path, ltype)
    return fields


def _do_infer_fields(coll, path, embedded_doc_type):
    fields = _DEFAULT_LABEL_FIELDS.get(embedded_doc_type, [])
    return [_make_field_doc(*f) for f in fields]


def _make_field_doc(name, ftype, subfield):
    if ftype == "fiftyone.core.fields.ObjectIdField":
        db_field = "_" + name
    else:
        db_field = name

    return {
        "_cls": "SampleFieldDocument",
        "name": name,
        "ftype": ftype,
        "subfield": subfield,
        "embedded_doc_type": None,
        "db_field": db_field,
        "fields": [],
    }


# format: (name, ftype, subfield)
_DEFAULT_LABEL_FIELDS = {
    "fiftyone.core.metadata.Metadata": [
        ("size_bytes", "fiftyone.core.fields.IntField", None),
        ("mime_type", "fiftyone.core.fields.StringField", None),
    ],
    "fiftyone.core.metadata.ImageMetadata": [
        ("size_bytes", "fiftyone.core.fields.IntField", None),
        ("mime_type", "fiftyone.core.fields.StringField", None),
        ("width", "fiftyone.core.fields.IntField", None),
        ("height", "fiftyone.core.fields.IntField", None),
        ("num_channels", "fiftyone.core.fields.IntField", None),
    ],
    "fiftyone.core.metadata.VideoMetadata": [
        ("size_bytes", "fiftyone.core.fields.IntField", None),
        ("mime_type", "fiftyone.core.fields.StringField", None),
        ("frame_width", "fiftyone.core.fields.IntField", None),
        ("frame_height", "fiftyone.core.fields.IntField", None),
        ("frame_rate", "fiftyone.core.fields.FloatField", None),
        ("total_frame_count", "fiftyone.core.fields.IntField", None),
        ("duration", "fiftyone.core.fields.FloatField", None),
        ("encoding_str", "fiftyone.core.fields.StringField", None),
    ],
    "fiftyone.core.labels.Regression": [
        ("id", "fiftyone.core.fields.ObjectIdField", None),
        (
            "tags",
            "fiftyone.core.fields.ListField",
            "fiftyone.core.fields.StringField",
        ),
        ("value", "fiftyone.core.fields.FloatField", None),
        ("confidence", "fiftyone.core.fields.FloatField", None),
    ],
    "fiftyone.core.labels.Classification": [
        ("id", "fiftyone.core.fields.ObjectIdField", None),
        (
            "tags",
            "fiftyone.core.fields.ListField",
            "fiftyone.core.fields.StringField",
        ),
        ("label", "fiftyone.core.fields.StringField", None),
        ("confidence", "fiftyone.core.fields.FloatField", None),
        ("logits", "fiftyone.core.fields.VectorField", None),
    ],
    "fiftyone.core.labels.Detection": [
        ("id", "fiftyone.core.fields.ObjectIdField", None),
        (
            "tags",
            "fiftyone.core.fields.ListField",
            "fiftyone.core.fields.StringField",
        ),
        ("label", "fiftyone.core.fields.StringField", None),
        (
            "bounding_box",
            "fiftyone.core.fields.ListField",
            "fiftyone.core.fields.FloatField",
        ),
        ("mask", "fiftyone.core.fields.ArrayField", None),
        ("confidence", "fiftyone.core.fields.FloatField", None),
        ("index", "fiftyone.core.fields.IntField", None),
    ],
    "fiftyone.core.labels.Polyline": [
        ("id", "fiftyone.core.fields.ObjectIdField", None),
        (
            "tags",
            "fiftyone.core.fields.ListField",
            "fiftyone.core.fields.StringField",
        ),
        ("label", "fiftyone.core.fields.StringField", None),
        ("points", "fiftyone.core.fields.PolylinePointsField", None),
        ("confidence", "fiftyone.core.fields.FloatField", None),
        ("index", "fiftyone.core.fields.IntField", None),
        ("closed", "fiftyone.core.fields.BooleanField", None),
        ("filled", "fiftyone.core.fields.BooleanField", None),
    ],
    "fiftyone.core.labels.Keypoint": [
        ("id", "fiftyone.core.fields.ObjectIdField", None),
        (
            "tags",
            "fiftyone.core.fields.ListField",
            "fiftyone.core.fields.StringField",
        ),
        ("label", "fiftyone.core.fields.StringField", None),
        ("points", "fiftyone.core.fields.PolylinePointsField", None),
        ("confidence", "fiftyone.core.fields.FloatField", None),
        ("index", "fiftyone.core.fields.IntField", None),
    ],
    "fiftyone.core.labels.Segmentation": [
        ("id", "fiftyone.core.fields.ObjectIdField", None),
        (
            "tags",
            "fiftyone.core.fields.ListField",
            "fiftyone.core.fields.StringField",
        ),
        ("mask", "fiftyone.core.fields.ArrayField", None),
    ],
    "fiftyone.core.labels.Heatmap": [
        ("id", "fiftyone.core.fields.ObjectIdField", None),
        (
            "tags",
            "fiftyone.core.fields.ListField",
            "fiftyone.core.fields.StringField",
        ),
        ("map", "fiftyone.core.fields.ArrayField", None),
        ("range", "fiftyone.core.fields.HeatmapRangeField", None),
    ],
    "fiftyone.core.labels.TemporalDetection": [
        ("id", "fiftyone.core.fields.ObjectIdField", None),
        (
            "tags",
            "fiftyone.core.fields.ListField",
            "fiftyone.core.fields.StringField",
        ),
        ("label", "fiftyone.core.fields.StringField", None),
        ("support", "fiftyone.core.fields.FrameSupportField", None),
        ("confidence", "fiftyone.core.fields.FloatField", None),
    ],
    "fiftyone.core.labels.GeoLocation": [
        ("id", "fiftyone.core.fields.ObjectIdField", None),
        (
            "tags",
            "fiftyone.core.fields.ListField",
            "fiftyone.core.fields.StringField",
        ),
        ("point", "fiftyone.core.fields.GeoPointField", None),
        ("line", "fiftyone.core.fields.GeoLineStringField", None),
        ("polygon", "fiftyone.core.fields.GeoPolygonField", None),
    ],
    "fiftyone.core.labels.GeoLocations": [
        ("id", "fiftyone.core.fields.ObjectIdField", None),
        (
            "tags",
            "fiftyone.core.fields.ListField",
            "fiftyone.core.fields.StringField",
        ),
        ("points", "fiftyone.core.fields.GeoMultiPointField", None),
        ("lines", "fiftyone.core.fields.GeoMultiLineStringField", None),
        ("polygons", "fiftyone.core.fields.GeoMultiPolygonField", None),
    ],
}

_LABEL_LIST_FIELDS = {
    "fiftyone.core.labels.Classifications": "classifications",
    "fiftyone.core.labels.Detections": "detections",
    "fiftyone.core.labels.Keypoints": "keypoints",
    "fiftyone.core.labels.Polylines": "polylines",
    "fiftyone.core.labels.TemporalDetections": "detections",
}



"""
FiftyOne migrations runner.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import bisect
import logging
import os
from packaging.requirements import Requirement
from packaging.version import Version as V

import eta.core.utils as etau

import fiftyone as fo
import fiftyone.constants as foc
import fiftyone.core.odm as foo
import fiftyone.core.utils as fou


logger = logging.getLogger(__name__)


DOWN = "down"
UP = "up"


def Version(version):
    """A ``packaging.version.Version`` proxy that excludes things like RC
    version info.
    """
    return V(V(version).base_version)


def _migrations_disabled():
    return os.environ.get("FIFTYONE_DISABLE_MIGRATIONS", "0") != "0"


def get_database_revision():
    """Gets the current revision of the database.

    Returns:
        the database revision string
    """
    config = foo.get_db_config()
    return config.version


def get_dataset_revision(name):
    """Gets the current revision of the given dataset.

    Args:
        name: the name of the dataset

    Returns:
        the dataset revision string
    """
    conn = foo.get_db_conn()
    dataset_doc = conn.datasets.find_one({"name": name}, {"version": 1})
    if dataset_doc is None:
        raise ValueError("Dataset '%s' not found" % name)

    return dataset_doc.get("version", None)


def migrate_all(destination=None, error_level=0, verbose=False):
    """Migrates the database and all datasets to the specified destination
    revision.

    If ``fiftyone.config.database_admin`` is ``False`` and no ``destination``
    is provided, the database and each dataset will only be migrated if their
    current versions are not compatible with the client's version.

    Args:
        destination (None): the destination revision. By default, the
            ``fiftyone`` client version is used
        error_level (0): the error level to use if an individual dataset cannot
            be migrated. Valid values are:

            -   0: raise error if a dataset cannot be migrated
            -   1: log warning if a dataset cannot be migrated
            -   2: ignore datasets that cannot be migrated
        verbose (False): whether to log incremental migrations that are run
    """
    migrate_database_if_necessary(destination=destination, verbose=verbose)

    for name in fo.list_datasets():
        migrate_dataset_if_necessary(
            name,
            destination=destination,
            error_level=error_level,
            verbose=verbose,
        )


def migrate_database_if_necessary(destination=None, verbose=False):
    """Migrates the database to the specified revision, if necessary.

    If ``fiftyone.config.database_admin`` is ``False`` and no ``destination``
    is provided, the database will only be migrated if its current version is
    not compatible with the client's version.

    Args:
        destination (None): the destination revision. By default, the
            ``fiftyone`` client version is used
        verbose (False): whether to log incremental migrations that are run
    """
    if _migrations_disabled():
        return

    config = foo.get_db_config()
    head = config.version

    default_destination = destination is None

    if default_destination:
        if not fo.config.database_admin and _is_compatible_version(head):
            return

        destination = foc.VERSION

    if head == destination:
        return

    if not fo.config.database_admin:
        if default_destination:
            if foc.COMPATIBLE_VERSIONS:
                compat_str = " (compatibility %s)" % foc.COMPATIBLE_VERSIONS
            else:
                compat_str = ""

            raise EnvironmentError(
                "Cannot connect to database v%s with client v%s%s. "
                "See https://voxel51.com/docs/fiftyone/user_guide/config.html#database-migrations "
                "for more information" % (head, destination, compat_str)
            )
        else:
            raise EnvironmentError(
                "Cannot migrate database from v%s to v%s when database_admin=%s. "
                "See https://voxel51.com/docs/fiftyone/user_guide/config.html#database-migrations "
                "for more information"
                % (head, destination, fo.config.database_admin)
            )

    if _database_exists():
        runner = MigrationRunner(head, destination)
        if runner.has_admin_revisions:
            logger.info("Migrating database to v%s", destination)
            runner.run_admin(verbose=verbose)

    config.version = destination
    config.save()


def needs_migration(name=None, head=None, destination=None):
    """Determines whether a dataset requires a migration in order to be used in
    the specified destination revision.

    To use this method, specify either the ``name`` of an existing dataset or
    provide the ``head`` revision of the dataset.

    If ``fiftyone.config.database_admin`` is ``False`` and no ``destination``
    is provided, a dataset will be deemed to require no migration if its
    current version if compatible with the client's version.

    Args:
        name (None): the name of the dataset
        head (None): the current revision of the dataset
        destination (None): the destination revision. By default, the current
            database version is used

    Returns:
        True/False
    """
    if name is not None:
        head = get_dataset_revision(name)

    if head is None:
        head = "0.0"

    if destination is None:
        if not fo.config.database_admin and _is_compatible_version(head):
            return False

        destination = get_database_revision()

    if head == destination:
        return False

    runner = MigrationRunner(head, destination)
    return runner.has_revisions


def migrate_dataset_if_necessary(
    name,
    destination=None,
    error_level=0,
    verbose=False,
):
    """Migrates the dataset from its current revision to the specified
    destination revision.

    If ``fiftyone.config.database_admin`` is ``False`` and no ``destination``
    is provided, the dataset will only be migrated if its current version is
    not compatible with the client's version.

    Args:
        name: the name of the dataset
        destination (None): the destination revision. By default, the current
            database version is used
        error_level (0): the error level to use. Valid values are:

            -   0: raise error if the dataset cannot be migrated
            -   1: log warning if the dataset cannot be migrated
            -   2: ignore datasets that cannot be migrated
        verbose (False): whether to log incremental migrations that are run
    """
    try:
        _migrate_dataset_if_necessary(name, destination, verbose)
    except Exception as e:
        fou.handle_error(e, error_level=error_level)


def _migrate_dataset_if_necessary(name, destination, verbose):
    if _migrations_disabled():
        return

    head = get_dataset_revision(name)
    db_version = get_database_revision()

    if head is None:
        head = "0.0"

    default_destination = destination is None

    if default_destination:
        if not fo.config.database_admin and _is_compatible_version(head):
            return

        destination = db_version

    if head == destination:
        return

    if not fo.config.database_admin and destination != db_version:
        if default_destination:
            if foc.COMPATIBLE_VERSIONS:
                compat_str = " (compatibility %s)" % foc.COMPATIBLE_VERSIONS
            else:
                compat_str = ""

            raise EnvironmentError(
                "Cannot load dataset '%s' from v%s with client v%s%s. "
                "See https://voxel51.com/docs/fiftyone/user_guide/config.html#database-migrations "
                "for more information" % (name, head, foc.VERSION, compat_str)
            )
        else:
            raise EnvironmentError(
                "Cannot migrate dataset '%s' from v%s to v%s when database_admin=%s."
                "See https://voxel51.com/docs/fiftyone/user_guide/config.html#database-migrations "
                "for more information"
                % (name, head, destination, fo.config.database_admin)
            )

    runner = MigrationRunner(head, destination)
    if runner.has_revisions:
        logger.info("Migrating dataset '%s' to v%s", name, destination)
        runner.run(name, verbose=verbose)

    if Version(destination) >= Version("0.6.2"):
        conn = foo.get_db_conn()
        conn.datasets.update_one(
            {"name": name}, {"$set": {"version": destination}}
        )
    else:
        # Old version of FiftyOne that didn't store dataset versions
        pass


class MigrationRunner(object):
    """Class for running FiftyOne migrations.

    Args:
        head: the current revision
        destination: the destination revision
    """

    def __init__(
        self,
        head,
        destination,
        _revisions=None,
        _admin_revisions=None,
    ):
        pkg_ver = Version(foc.VERSION)
        head_ver = Version(head)
        dest_ver = Version(destination)
        if head_ver > pkg_ver or dest_ver > pkg_ver:
            raise EnvironmentError(
                "You must have fiftyone>=%s installed in order to migrate "
                "from v%s to v%s, but you are currently running fiftyone==%s."
                "\n\nSee https://voxel51.com/docs/fiftyone/getting_started/install.html#downgrading-fiftyone "
                "for information about downgrading FiftyOne."
                % (max(head_ver, dest_ver), head_ver, dest_ver, pkg_ver)
            )

        if _revisions is None:
            _revisions = _get_all_revisions()

        if _admin_revisions is None:
            _admin_revisions = _get_all_revisions(admin=True)

        self._head = head
        self._destination = destination
        self._revisions, self._direction = _get_revisions_to_run(
            head, destination, _revisions
        )
        self._admin_revisions, _ = _get_revisions_to_run(
            head, destination, _admin_revisions
        )

    @property
    def head(self):
        """The head revision."""
        return self._head

    @property
    def destination(self):
        """The destination revision."""
        return self._destination

    @property
    def direction(self):
        """The direction of the migration runner; one of ``("up", "down").``"""
        return self._direction

    @property
    def has_revisions(self):
        """Whether there are any revisions to run."""
        return bool(self._revisions)

    @property
    def has_admin_revisions(self):
        """Whether there are any admin revisions to run."""
        return bool(self._admin_revisions)

    @property
    def revisions(self):
        """The list of revisions that will be run by :meth:`run`."""
        return [r[0] for r in self._revisions]

    @property
    def admin_revisions(self):
        """The list of admin revisions that will be run by :meth:`run_admin`."""
        return [r[0] for r in self._admin_revisions]

    def run(self, dataset_name, verbose=False):
        """Runs any required migrations on the specified dataset.

        Args:
            dataset_name: the name of the dataset to migrate
            verbose (False): whether to log incremental migrations that are run
        """
        conn = foo.get_db_conn()
        for rev, module in self._revisions:
            if verbose:
                logger.info("Running v%s %s migration", rev, self.direction)

            fcn = etau.get_function(self.direction, module)
            fcn(conn, dataset_name)

    def run_admin(self, verbose=False):
        """Runs any required admin revisions.

        Args:
            verbose (False): whether to log incremental migrations that are run
        """
        client = foo.get_db_client()
        for rev, module in self._admin_revisions:
            if verbose:
                logger.info(
                    "Running v%s %s admin migration", rev, self.direction
                )

            fcn = etau.get_function(self.direction, module)
            fcn(client)


def _database_exists():
    client = foo.get_db_client()
    return fo.config.database_name in client.list_database_names()


def _is_compatible_version(version):
    specifier = foc.COMPATIBLE_VERSIONS or ("==%s" % foc.VERSION)
    req = Requirement("fiftyone" + specifier)
    return req.specifier.contains(version)


def _get_revisions_to_run(head, dest, revisions):
    revisions = sorted(revisions, key=lambda r: Version(r[0]))

    head = Version(head)
    dest = Version(dest)

    rev_versions = [Version(r[0]) for r in revisions]

    head_idx = bisect.bisect(rev_versions, head)
    dest_idx = bisect.bisect(rev_versions, dest)

    if dest >= head:
        revisions_to_run = revisions[head_idx:dest_idx]
        return revisions_to_run, UP

    revisions_to_run = revisions[dest_idx:head_idx][::-1]
    return revisions_to_run, DOWN


def _get_all_revisions(admin=False):
    revisions_dir = foc.MIGRATIONS_REVISIONS_DIR
    if admin:
        revisions_dir = os.path.join(revisions_dir, "admin")

    revision_files = [
        f for f in etau.list_files(revisions_dir) if f.endswith(".py")
    ]

    module_prefix = __loader__.name.rsplit(".", 1)[0] + ".revisions"
    if admin:
        module_prefix = module_prefix + ".admin"

    revisions = []
    for filename in revision_files:
        version = filename[1:-3].replace("_", ".")
        module = module_prefix + "." + filename[:-3]
        revisions.append((version, module))

    return sorted(revisions, key=lambda r: Version(r[0]))



"""
Notebook Session HTML templates

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from jinja2 import Template


SCREENSHOT_STYLE = """
@import url("https://fonts.googleapis.com/css2?family=Palanquin&display=swap");

#focontainer-{{ subscription }} {
  position: relative;
  height: {{ height }}px;
  display: block !important;
}
#foactivate-{{ subscription }} {
  font-weight: bold;
  cursor: pointer;
  font-size: 24px;
  border-radius: 3px;
  text-align: center;
  padding: 0.5em;
  color: rgb(255, 255, 255);
  font-family: "Palanquin", sans-serif;
  position: absolute;
  left: 50%;
  top: 50%;
  width: 160px;
  margin-left: -80px;
  margin-top: -23px;
  background: hsla(210,11%,15%, 0.8);
  border: none;
}
#foactivate-{{ subscription }}:focus {
  outline: none;
}
#fooverlay-{{ subscription }} {
  width: 100%;
  height: 100%;
  background: hsla(208, 7%, 46%, 0.7);
  position: absolute;
  top: 0;
  left: 0;
  display: none;
  cursor: pointer;
}
"""

SCREENSHOT_SCRIPT = """
   (function() {
     var container = document.getElementById("focontainer-{{ subscription }}");
     var overlay = document.getElementById("fooverlay-{{ subscription }}");
     fetch(`{{ url }}fiftyone`)
     .then(() => {
        overlay.addEventListener("click", () => {
          fetch(`{{ url }}event`, {
            method: "POST",
            body: JSON.stringify({
              event: "reactivate_notebook_cell",
              data: { subscription: "{{ subscription }}" },
              subscription: "{{ subscription }}"
            })
          })
        });
        container.addEventListener("mouseenter", () => overlay.style.display = "block");
        container.addEventListener("mouseleave", () => overlay.style.display = "none");
     });
   })();
"""


SCREENSHOT_DIV = """
<div id="focontainer-{{ subscription }}">
   <div id="fooverlay-{{ subscription }}" style="display: none;">
      <button id="foactivate-{{ subscription }}" >Activate</button>
   </div>
   <img src='{{ image }}' style="width: 100%; max-width: {{ max_width }}px;"/>
</div>
"""


SCREENSHOT_HTML = Template(
    """
<style>%s</style>
%s
<script type="text/javascript">%s</script>
"""
    % (SCREENSHOT_STYLE, SCREENSHOT_DIV, SCREENSHOT_SCRIPT)
)


SCREENSHOT_COLAB = """
<style>
{{ style }}
</style>
<div id="focontainer-{{ subscription }}" style="display: none;">
   <div id="fooverlay-{{ subscription }}">
      <button id="foactivate-{{ subscription }}" >Activate</button>
   </div>
</div>
"""


SCREENSHOT_COLAB_SCRIPT = """
(() => {
    google.colab.kernel.proxyPort({{ port }}, {
        'cache': true
    }).then((baseURL) => {
        const url = new URL(baseURL);
        url.searchParams.set('context', 'colab');
        url.searchParams.set('polling', 'true');
        const subscription = "{{ subscription }}";
        url.searchParams.set('subscription', subscription);
        const iframe = document.createElement('iframe');
        iframe.src = url;
        iframe.setAttribute('width', '100%');
        iframe.setAttribute('height', '{{ height }}');
        iframe.setAttribute('frameborder', 0);
        document.body.appendChild(iframe);
        window.addEventListener("message", (event) => {
            if (event.data.subscription !== subscription) return;
            document.body.removeChild(iframe);
            var container = document.getElementById(`focontainer-${subscription}`);
            var overlay = document.getElementById(`fooverlay-${subscription}`);
            google.colab.kernel.invokeFunction(`fiftyone.${subscription.replaceAll('-', '_')}`, [event.data.src, event.data.width], {});
            overlay.addEventListener("click", () => {
              google.colab.kernel.invokeFunction(`fiftyone.deactivate.${subscription.replaceAll('-', '_')}`, [], {})
              container.removeChild(container.children[1]);
              document.body.appendChild(iframe);
            });
            container.addEventListener("mouseenter", () => overlay.style.display = "block");
            container.addEventListener("mouseleave", () => overlay.style.display = "none");
        });
    });
})()
"""



"""
Base evaluation methods.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import itertools

import numpy as np
import sklearn.metrics as skm

import fiftyone.core.evaluation as foe
import fiftyone.core.plots as fop


class BaseEvaluationResults(foe.EvaluationResults):
    """Base class for evaluation results.

    Args:
        ytrue: a list of ground truth labels
        ypred: a list of predicted labels
        confs (None): an optional list of confidences for the predictions
        weights (None): an optional list of sample weights
        eval_key (None): the evaluation key of the evaluation
        gt_field (None): the name of the ground truth field
        pred_field (None): the name of the predictions field
        ytrue_ids (None): a list of IDs for the ground truth labels
        ypred_ids (None): a list of IDs for the predicted labels
        classes (None): the list of possible classes. If not provided, the
            observed ground truth/predicted labels are used
        missing (None): a missing label string. Any None-valued labels are
            given this label for evaluation purposes
        samples (None): the :class:`fiftyone.core.collections.SampleCollection`
            for which the results were computed
    """

    def __init__(
        self,
        ytrue,
        ypred,
        confs=None,
        weights=None,
        eval_key=None,
        gt_field=None,
        pred_field=None,
        ytrue_ids=None,
        ypred_ids=None,
        classes=None,
        missing=None,
        samples=None,
    ):
        if missing is None:
            missing = "(none)"

        ytrue, ypred, classes = _parse_labels(ytrue, ypred, classes, missing)

        self.ytrue = np.asarray(ytrue)
        self.ypred = np.asarray(ypred)
        self.confs = np.asarray(confs) if confs is not None else None
        self.weights = np.asarray(weights) if weights is not None else None
        self.eval_key = eval_key
        self.gt_field = gt_field
        self.pred_field = pred_field
        self.ytrue_ids = (
            np.asarray(ytrue_ids) if ytrue_ids is not None else None
        )
        self.ypred_ids = (
            np.asarray(ypred_ids) if ypred_ids is not None else None
        )
        self.classes = np.asarray(classes)
        self.missing = missing

        self._samples = samples

    def report(self, classes=None):
        """Generates a classification report for the results via
        :func:`sklearn:sklearn.metrics.classification_report`.

        Args:
            classes (None): an optional list of classes to include in the
                report

        Returns:
            a dict
        """
        labels = self._parse_classes(classes)

        if self.ytrue.size == 0 or labels.size == 0:
            d = {}
            empty = {
                "precision": 0.0,
                "recall": 0.0,
                "f1-score": 0.0,
                "support": 0,
            }

            if labels.size > 0:
                for label in labels:
                    d[label] = empty.copy()

            d["micro avg"] = empty.copy()
            d["macro avg"] = empty.copy()
            d["weighted avg"] = empty.copy()
            return d

        return skm.classification_report(
            self.ytrue,
            self.ypred,
            labels=labels,
            sample_weight=self.weights,
            output_dict=True,
            zero_division=0,
        )

    def metrics(self, classes=None, average="micro", beta=1.0):
        """Computes classification metrics for the results, including accuracy,
        precision, recall, and F-beta score.

        See :func:`sklearn:sklearn.metrics.precision_recall_fscore_support` for
        details.

        Args:
            classes (None): an optional list of classes to include in the
                calculations
            average ("micro"): the averaging strategy to use
            beta (1.0): the F-beta value to use

        Returns:
            a dict
        """
        labels = self._parse_classes(classes)

        accuracy = _compute_accuracy(
            self.ytrue, self.ypred, labels=labels, weights=self.weights
        )

        precision, recall, fscore, _ = skm.precision_recall_fscore_support(
            self.ytrue,
            self.ypred,
            average=average,
            labels=labels,
            beta=beta,
            sample_weight=self.weights,
            zero_division=0,
        )

        support = _compute_support(
            self.ytrue, labels=labels, weights=self.weights
        )

        return {
            "accuracy": accuracy,
            "precision": precision,
            "recall": recall,
            "fscore": fscore,
            "support": support,
        }

    def print_report(self, classes=None, digits=2):
        """Prints a classification report for the results via
        :func:`sklearn:sklearn.metrics.classification_report`.

        Args:
            classes (None): an optional list of classes to include in the
                report
            digits (2): the number of digits of precision to print
        """
        labels = self._parse_classes(classes)

        if labels.size == 0:
            print("No classes to analyze")
            return

        report_str = skm.classification_report(
            self.ytrue,
            self.ypred,
            labels=labels,
            digits=digits,
            sample_weight=self.weights,
            zero_division=0,
        )
        print(report_str)

    def confusion_matrix(self, classes=None, include_other=False):
        """Generates a confusion matrix for the results via
        :func:`sklearn:sklearn.metrics.confusion_matrix`.

        The rows of the confusion matrix represent ground truth and the columns
        represent predictions.

        Args:
            classes (None): an optional list of classes to include in the
                confusion matrix. Include ``self.missing`` in this list if you
                would like to include a row/column for unmatched examples
            include_other (False): whether to include an extra row/column at
                the end of the matrix for labels that do not appear in
                ``classes``. Only applicable if ``classes`` are provided

        Returns:
            a ``num_classes x num_classes`` confusion matrix
        """
        confusion_matrix, _, _ = self._confusion_matrix(
            classes=classes, include_other=include_other, include_missing=False
        )
        return confusion_matrix

    def plot_confusion_matrix(
        self,
        classes=None,
        include_other=None,
        include_missing=None,
        other_label="(other)",
        backend="plotly",
        **kwargs,
    ):
        """Plots a confusion matrix for the evaluation results.

        If you are working in a notebook environment with the default plotly
        backend, this method returns an interactive
        :class:`fiftyone.core.plots.plotly.InteractiveHeatmap` that you can
        attach to an App session via its
        :attr:`fiftyone.core.session.Session.plots` attribute, which will
        automatically sync the session's view with the currently selected cells
        in the confusion matrix.

        Args:
            classes (None): an optional list of classes to include in the
                confusion matrix
            include_other (None): whether to include a row/column for examples
                whose label is in ``classes`` but are matched to labels that
                do not appear in ``classes``. Only applicable if ``classes``
                are provided. The supported values are:

                -   None (default): only include a row/column for other labels
                    if there are any
                -   True: do include a row/column for other labels
                -   False: do not include a row/column for other labels
            include_missing (None): whether to include a row/column for missing
                ground truth/predictions in the confusion matrix. The supported
                values are:

                -   None (default): only include a row/column for missing
                    labels if there are any
                -   True: do include a row/column for missing labels
                -   False: do not include a row/column for missing labels
            other_label ("(other)"): the label to use for "other" predictions
            backend ("plotly"): the plotting backend to use. Supported values
                are ``("plotly", "matplotlib")``
            **kwargs: keyword arguments for the backend plotting method:

                -   "plotly" backend: :meth:`fiftyone.core.plots.plotly.plot_confusion_matrix`
                -   "matplotlib" backend: :meth:`fiftyone.core.plots.matplotlib.plot_confusion_matrix`

        Returns:
            one of the following:

            -   a :class:`fiftyone.core.plots.plotly.InteractiveHeatmap`, if
                the plotly backend is used
            -   a matplotlib figure, otherwise
        """
        confusion_matrix, labels, ids = self._confusion_matrix(
            classes=classes,
            include_other=include_other,
            include_missing=include_missing,
            other_label=other_label,
            tabulate_ids=True,
        )

        return fop.plot_confusion_matrix(
            confusion_matrix,
            labels,
            ids=ids,
            samples=self._samples,
            eval_key=self.eval_key,
            gt_field=self.gt_field,
            pred_field=self.pred_field,
            backend=backend,
            **kwargs,
        )

    def _parse_classes(self, classes):
        if classes is not None:
            return np.asarray(classes)

        return np.array([c for c in self.classes if c != self.missing])

    def _confusion_matrix(
        self,
        classes=None,
        include_other=None,
        include_missing=None,
        other_label=None,
        tabulate_ids=False,
    ):
        if classes is not None:
            labels = list(classes)
        else:
            labels = list(self.classes)
            include_other = False

        if include_other != False and other_label not in labels:
            added_other = True
            labels.append(other_label)
        else:
            added_other = False

        if include_missing != False and self.missing not in labels:
            added_missing = True
            labels.append(self.missing)
        else:
            added_missing = False

        if include_other != False:
            labels_set = set(labels + [self.missing])
            ypred = [y if y in labels_set else other_label for y in self.ypred]
            ytrue = [y if y in labels_set else other_label for y in self.ytrue]
        else:
            ypred = self.ypred
            ytrue = self.ytrue

        cmat, ids = _compute_confusion_matrix(
            ytrue,
            ypred,
            labels,
            weights=self.weights,
            ytrue_ids=self.ytrue_ids,
            ypred_ids=self.ypred_ids,
            tabulate_ids=tabulate_ids,
        )

        if added_other:
            # Omit `(other, other)`
            i = labels.index(other_label)
            cmat[i, i] = 0
            ids[i, i] = []

            if added_missing:
                # Omit `(other, missing)` and `(missing, other)`
                j = labels.index(self.missing)
                cmat[i, j] = 0
                cmat[j, i] = 0
                ids[i, j] = []
                ids[j, i] = []

        rm_inds = []

        if include_other == None:
            idx = labels.index(other_label)
            if not any(cmat[:, idx]) and not any(cmat[idx, :]):
                rm_inds.append(idx)

        if include_missing == None:
            idx = labels.index(self.missing)
            if not any(cmat[:, idx]) and not any(cmat[idx, :]):
                rm_inds.append(idx)

        if rm_inds:
            cmat = np.delete(np.delete(cmat, rm_inds, axis=0), rm_inds, axis=1)
            ids = np.delete(np.delete(ids, rm_inds, axis=0), rm_inds, axis=1)
            labels = [l for i, l in enumerate(labels) if i not in rm_inds]

        return cmat, labels, ids

    @classmethod
    def _from_dict(cls, d, samples, config, **kwargs):
        ytrue = d["ytrue"]
        ypred = d["ypred"]
        confs = d.get("confs", None)
        weights = d.get("weights", None)
        eval_key = d.get("eval_key", None)
        gt_field = d.get("gt_field", None)
        pred_field = d.get("pred_field", None)
        ytrue_ids = d.get("ytrue_ids", None)
        ypred_ids = d.get("ypred_ids", None)
        classes = d.get("classes", None)
        missing = d.get("missing", None)
        return cls(
            ytrue,
            ypred,
            confs=confs,
            weights=weights,
            eval_key=eval_key,
            gt_field=gt_field,
            pred_field=pred_field,
            ytrue_ids=ytrue_ids,
            ypred_ids=ypred_ids,
            classes=classes,
            missing=missing,
            samples=samples,
            **kwargs,
        )


def _parse_labels(ytrue, ypred, classes, missing):
    if classes is None:
        classes = set(ytrue) | set(ypred)
        classes.discard(None)
        classes = sorted(classes)
    else:
        classes = list(classes)

    ytrue = [y if y is not None else missing for y in ytrue]
    ypred = [y if y is not None else missing for y in ypred]

    return ytrue, ypred, classes


def _compute_accuracy(ytrue, ypred, labels=None, weights=None):
    if labels is not None:
        labels = set(labels)
        found = np.array(
            [yt in labels or yp in labels for yt, yp in zip(ytrue, ypred)],
            dtype=bool,
        )
        ytrue = ytrue[found]
        ypred = ypred[found]
        if weights is not None:
            weights = weights[found]

    if ytrue.size > 0:
        scores = ytrue == ypred
        if weights is not None:
            try:
                accuracy = np.sum(weights * scores) / np.sum(weights)
            except ZeroDivisionError:
                accuracy = 0.0
        else:
            accuracy = np.mean(scores)
    else:
        accuracy = 0.0

    return accuracy


def _compute_support(ytrue, labels=None, weights=None):
    if labels is not None:
        labels = set(labels)
        found = np.array([y in labels for y in ytrue], dtype=bool)
        ytrue = ytrue[found]
        if weights is not None:
            weights = weights[found]

    if weights is not None:
        support = np.sum(weights)
    else:
        support = ytrue.size

    return support


def _compute_confusion_matrix(
    ytrue,
    ypred,
    labels,
    weights=None,
    ytrue_ids=None,
    ypred_ids=None,
    tabulate_ids=False,
):
    ytrue = np.asarray(ytrue).flatten()
    ypred = np.asarray(ypred).flatten()
    labels = np.asarray(labels)

    if weights is None:
        weights = np.ones(ytrue.size, dtype=int)
    else:
        weights = np.asarray(weights).flatten()

    if weights.dtype.kind in {"i", "u", "b"}:
        dtype = np.int64
    else:
        dtype = np.float64

    num_labels = labels.size

    confusion_matrix = np.zeros((num_labels, num_labels), dtype=dtype)

    if tabulate_ids:
        ids = np.empty((num_labels, num_labels), dtype=object)
        for i in range(num_labels):
            for j in range(num_labels):
                ids[i, j] = []
    else:
        ytrue_ids = None
        ypred_ids = None
        ids = None

    if num_labels == 0 or ytrue.size == 0:
        return confusion_matrix, ids

    labels_to_inds = {label: idx for idx, label in enumerate(labels)}
    ypred = np.array([labels_to_inds.get(y, -1) for y in ypred])
    ytrue = np.array([labels_to_inds.get(y, -1) for y in ytrue])

    found = np.logical_and(ypred >= 0, ytrue >= 0)
    ypred = ypred[found]
    ytrue = ytrue[found]
    weights = weights[found]

    if ytrue_ids is not None:
        ytrue_ids = ytrue_ids[found]
    else:
        ytrue_ids = itertools.repeat(None)

    if ypred_ids is not None:
        ypred_ids = ypred_ids[found]
    else:
        ypred_ids = itertools.repeat(None)

    for yt, yp, w, it, ip in zip(ytrue, ypred, weights, ytrue_ids, ypred_ids):
        confusion_matrix[yt, yp] += w

        if it is not None:
            ids[yt, yp].append(it)

        if ip is not None:
            ids[yt, yp].append(ip)

    return confusion_matrix, ids



"""
FiftyOne Server aggregations

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from datetime import date, datetime, timedelta
import typing as t

import asyncio
import strawberry as gql

import fiftyone as fo
import fiftyone.core.aggregations as foa
import fiftyone.core.collections as foc

from fiftyone.server.constants import LIST_LIMIT
from fiftyone.server.data import T
from fiftyone.server.scalars import BSONArray


_DEFAULT_NUM_HISTOGRAM_BINS = 25


@gql.type
class ValueCount(t.Generic[T]):
    key: t.Union[T, None]
    value: int


@gql.type
class CountValuesResponse(t.Generic[T]):
    values: t.List[ValueCount[T]]


@gql.type
class BoolCountValuesResponse(CountValuesResponse[bool]):
    values: t.List[ValueCount[bool]]


@gql.type
class IntCountValuesResponse(CountValuesResponse[int]):
    values: t.List[ValueCount[int]]


@gql.type
class StrCountValuesResponse(CountValuesResponse[str]):
    values: t.List[ValueCount[str]]


CountValuesResponses = gql.union(
    "CountValuesResponses",
    (BoolCountValuesResponse, IntCountValuesResponse, StrCountValuesResponse),
)
COUNT_VALUES_TYPES = {fo.BooleanField, fo.IntField, fo.StringField}


@gql.type
class HistogramValuesResponse(t.Generic[T]):
    counts: t.List[int]
    edges: t.List[T]
    other: int


@gql.type
class DatetimeHistogramValuesResponse(HistogramValuesResponse[datetime]):
    edges: t.List[datetime]


@gql.type
class FloatHistogramValuesResponse(HistogramValuesResponse[float]):
    edges: t.List[float]


@gql.type
class IntHistogramValuesResponse(HistogramValuesResponse[int]):
    edges: t.List[float]


@gql.input
class HistogramValues:
    field: str


@gql.input
class CountValues:
    field: str


@gql.input
class Aggregate:
    count_values: t.Optional[CountValues] = None
    histogram_values: t.Optional[HistogramValues] = None


HistogramValuesResponses = gql.union(
    "HistogramValuesResponses",
    (
        DatetimeHistogramValuesResponse,
        FloatHistogramValuesResponse,
        IntHistogramValuesResponse,
    ),
)

HISTOGRAM_VALUES_TYPES = {
    fo.DateField,
    fo.DateTimeField,
    fo.IntField,
    fo.FloatField,
}


@gql.type
class AggregateQuery:
    @gql.field
    async def aggregate(
        self,
        dataset_name: str,
        view: BSONArray,
        aggregations: t.List[Aggregate],
    ) -> t.List[
        gql.union(
            "AggregationResponses",
            (
                BoolCountValuesResponse,
                IntCountValuesResponse,
                StrCountValuesResponse,
                DatetimeHistogramValuesResponse,
                FloatHistogramValuesResponse,
                IntHistogramValuesResponse,
            ),
        )
    ]:
        view = await load_view(dataset_name, view)

        resolvers = []
        aggs = []
        for input in aggregations:
            if input.count_values:
                resolve, agg = await _count_values(view, input.count_values)
            elif input.histogram_values:
                resolve, agg = await _histogram_values(
                    view, input.histogram_values
                )

            aggs.append(agg)
            resolvers.append(resolve)

        results = await view._async_aggregate(aggs)

        responses = []
        for resolver, result in zip(resolvers, results):
            responses.append(resolver(result))

        return responses


async def load_view(
    name: str, serialized_view: BSONArray
) -> foc.SampleCollection:
    def run() -> foc.SampleCollection:
        dataset = fo.load_dataset(name)
        dataset.reload()
        return fo.DatasetView._build(dataset, serialized_view or [])

    loop = asyncio.get_running_loop()

    return await loop.run_in_executor(None, run)


async def _count_values(
    view: foc.SampleCollection, input: CountValues
) -> t.Tuple[t.Callable[[t.List], CountValuesResponses], foa.CountValues]:
    field = view.get_field(input.field)

    while isinstance(field, fo.ListField):
        field = field.field

    def resolve(data: t.List):
        _, data = data
        values = [ValueCount(key=value, value=count) for value, count in data]

        if isinstance(field, fo.StringField):
            return StrCountValuesResponse(values=values)

        if isinstance(field, fo.BooleanField):
            return BoolCountValuesResponse(values=values)

        if isinstance(field, fo.IntField):
            return IntCountValuesResponse(values=values)

    return resolve, foa.CountValues(input.field, _first=LIST_LIMIT, _asc=False)


async def _histogram_values(
    view: foc.SampleCollection, input: HistogramValues
) -> t.Tuple[t.Callable[[t.List], HistogramValuesResponses], foa.CountValues]:
    field = view.get_field(input.field)

    while isinstance(field, fo.ListField):
        field = field.field

    range = await view._async_aggregate(
        foa.Bounds(input.field, safe=True, _count_nonfinites=True)
    )
    range_ = range.pop("bounds")
    bins = _DEFAULT_NUM_HISTOGRAM_BINS

    if isinstance(field, fo.IntField):
        if range_[0] is None:
            range_ = (0, 0)
        delta = range_[1] - range_[0]
        range_ = (range_[0] - 1, range_[1] + 1)
        if delta < _DEFAULT_NUM_HISTOGRAM_BINS:
            bins = delta + 1

    if range_[0] == range_[1]:
        bins = 1
        if range_[0] is None:
            range_ = [0, 1]

    if isinstance(range_[1], datetime):
        range_ = (range_[0], range_[1] + timedelta(milliseconds=100))
    elif isinstance(range_[1], date):
        range_ = (range_[0], range_[1] + timedelta(days=1))
    elif not isinstance(field, fo.IntField):
        range_ = (range_[0], range_[1] + 1e-6)

    def resolve(data):
        counts, edges, other = data
        data = {"counts": counts, "edges": edges, "other": other}
        if isinstance(field, (fo.DateField, fo.DateTimeField)):
            return DatetimeHistogramValuesResponse(**data)

        if isinstance(field, fo.FloatField):
            return FloatHistogramValuesResponse(**data)

        if isinstance(field, fo.IntField):
            return IntHistogramValuesResponse(**data)

    return resolve, foa.HistogramValues(input.field, bins=bins, range=range_)



"""
Session plot manager.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import datetime
import itertools
import logging

from bson import ObjectId

import eta.core.utils as etau

import fiftyone.core.clips as foc
from fiftyone.core.expressions import ViewField as F
import fiftyone.core.patches as fop
import fiftyone.core.video as fov

from .base import ResponsivePlot, ViewPlot, InteractivePlot


logger = logging.getLogger(__name__)


class PlotManager(object):
    """Class that manages communication between a
    :class:`fiftyone.core.session.Session` and one or more
    :class:`fiftyone.core.plots.base.ResponsivePlot` instances.

    Each plot can be linked to either the view, samples, frames, or labels of a
    session:

    -   **View links:** When a plot has ``link_type == "view"``, then, when the
        session's view changes, the plot is updated based on the content of the
        view

    -   **Sample links:** When points are selected in a plot with
        ``link_type == "samples"``, a view containing the corresponding samples
        is loaded in the App. Conversely, when the session's view changes, the
        corresponding points are selected in the plot

    -   **Frame links:** When points are selected in a plot with
        ``link_type == "frames"``, a view containing the corresponding frames
        is loaded in the App. Conversely, when the session's view changes, the
        corresponding points are selected in the plot

    -   **Label links:** When points are selected in a plot with
        ``link_type == "labels"``, a view containing the corresponding labels
        is loaded in the App. Conversely, when the session's view changes, the
        points in the plot corresponding to all labels in the view are selected
        in the plot

    Args:
        session: a :class:`fiftyone.core.session.Session`
    """

    _MIN_UPDATE_DELTA_SECONDS = 1

    def __init__(self, session):
        self._session = None
        self._plots = {}
        self._aggs = {}
        self._current_sample_ids = None
        self._current_frame_ids = None
        self._current_labels = None
        self._last_session_view = None
        self._last_update = None
        self._last_updates = {}
        self._connected = False
        self._disconnected = False

        self._set_session(session)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return self.summary()

    def __iter__(self):
        return iter(self._plots)

    def __contains__(self, name):
        return name in self._plots

    def __setitem__(self, name, plot):
        self.attach(plot, name=name)

    def __getitem__(self, name):
        return self._plots[name]

    def __delitem__(self, name):
        self.remove(name)

    def __bool__(self):
        return bool(self._plots)

    def __len__(self):
        return len(self._plots)

    def _set_session(self, session):
        if self.is_connected:
            self.disconnect()

        self._session = session

        if self._plots:
            self.connect()

    def summary(self):
        """Returns a string summary of this manager.

        Returns:
            a string summary
        """
        if not self._plots:
            return "No plots"

        elements = []

        connected_plots = [
            name for name, p in self._plots.items() if p.is_connected
        ]
        if connected_plots:
            elements.append("Connected plots:")
            elements.extend(self._summarize_plots(connected_plots))

        disconnected_plots = [
            name for name, p in self._plots.items() if not p.is_connected
        ]
        if disconnected_plots:
            elements.append("Disconnected plots:")
            elements.extend(self._summarize_plots(disconnected_plots))

        return "\n".join(elements)

    def _summarize_plots(self, names):
        maxlen = max(len(name) for name in names) + 1
        fmt = "    %%-%ds %%s" % maxlen

        elements = []
        for name in names:
            plot = self._plots[name]
            elements.append(fmt % (name + ":", etau.get_class_name(plot)))

        return elements

    def keys(self):
        """Returns an iterator over the names of plots in this manager.

        Returns:
            an interator over plot names
        """
        return self._plots.keys()

    def items(self):
        """Returns an iterator over the ``(name, plot)`` pairs in this manager.

        Returns:
            an iterator that emits ``(name, ResponsivePlot)`` tuples
        """
        return self._plots.items()

    def values(self):
        """Returns an iterator over the plots in this manager.

        Returns:
            an iterator that emits
            :class:`fiftyone.core.plots.base.ResponsivePlot` instances
        """
        return self._plots.values()

    @property
    def is_connected(self):
        """Whether this manager is currently connected to its plots."""
        return self._connected

    @property
    def is_disconnected(self):
        """Whether this manager is currently disconnected from its plots."""
        return self._disconnected

    @property
    def has_view_links(self):
        """Whether this manager has plots linked to views."""
        return any(
            plot.link_type == "view"
            for plot in self._plots.values()
            if plot.is_connected
        )

    @property
    def has_sample_links(self):
        """Whether this manager has plots linked to samples."""
        return any(
            plot.link_type == "samples"
            for plot in self._plots.values()
            if plot.is_connected
        )

    @property
    def has_frame_links(self):
        """Whether this manager has plots linked to frames."""
        return any(
            plot.link_type == "frames"
            for plot in self._plots.values()
            if plot.is_connected
        )

    @property
    def has_label_links(self):
        """Whether this manager has plots linked to labels."""
        return any(
            plot.link_type == "labels"
            for plot in self._plots.values()
            if plot.is_connected
        )

    def attach(self, plot, name=None, connect=True, overwrite=True):
        """Attaches a plot to this manager.

        Args:
            plot: a :class:`fiftyone.core.plots.base.ResponsivePlot`
            name (None): an optional name for the plot
            connect (True): whether to immediately connect the plot
            overwrite (True): whether to overwrite an existing plot of the same
                name
        """
        if not isinstance(plot, ResponsivePlot):
            raise ValueError(
                "Plots must be subclasses of %s; but found %s. You may be "
                "working in an environment that does not support "
                "interactivity.\n\nSee "
                "https://voxel51.com/docs/fiftyone/user_guide/plots.html#overview "
                "for more information" % (ResponsivePlot, type(plot))
            )

        same_plots = [(n, p) for n, p in self._plots.items() if p is plot]
        if same_plots:
            current_name = same_plots[0][0]
            if name is not None and name != current_name:
                logger.warning(
                    "Plot is already attached under name '%s'", current_name
                )

            return

        if name is None:
            name = "plot%d" % (len(self._plots) + 1)

        if name == "session":
            raise ValueError("Cannot use reserved name 'session' for a plot")

        if name in self._plots:
            if not overwrite:
                raise ValueError("A plot with name '%s' already exists" % name)

            _plot = self.pop(name)
            if _plot.is_connected:
                _plot.disconnect()

        self._plots[name] = plot

        if connect:
            if not self.is_connected:
                self.connect()
            else:
                self._connect_plot(name)
                self._update_plots([name])

    def remove(self, name):
        """Removes the plot from this manager.

        Args:
            name: the name of a plot
        """
        self.pop(name)

    def clear(self):
        """Removes all plots from this manager."""
        plot_names = list(self._plots.keys())
        for name in plot_names:
            self.remove(name)

    def pop(self, name):
        """Removes the plot from this manager and returns it.

        Args:
            name: the name of a plot

        Returns:
            the :class:`fiftyone.core.plots.base.ResponsivePlot`
        """
        if name not in self._plots:
            raise ValueError("No plot with name '%s'" % name)

        plot = self._plots.pop(name)
        self._aggs.pop(name, None)

        plot.disconnect()

        if not self._plots:
            self.disconnect()

        return plot

    def connect(self):
        """Connects this manager to its session and all plots."""
        if self.is_connected:
            return

        for name in self._plots:
            self._connect_plot(name)

        self._session._client.add_event_listener(
            "state_update", self._on_session_update
        )

        self._connected = True
        self._disconnected = False

        self.sync()

    def _connect_plot(self, name):
        plot = self._plots[name]

        if isinstance(plot, ViewPlot):
            # Record aggregations for efficient batch-updates
            self._aggs[name] = plot._get_aggregations()

            # Load current view
            plot.update_view(self._session._collection.view())

        if isinstance(plot, InteractivePlot):
            # Register plot's callbacks

            def _on_plot_selection(ids):
                self._on_plot_selection(name, ids)

            plot.register_selection_callback(_on_plot_selection)
            plot.register_sync_callback(self.sync)
            plot.register_disconnect_callback(self.disconnect)

        plot.connect()

    def disconnect(self):
        """Disconnects this manager from its session and all plots."""
        if not self.is_connected:
            return

        self._session._client.remove_event_listener(
            "state_update", self._on_session_update
        )

        for name in self._plots:
            self._disconnect_plot(name)

        self._last_update = None
        self._last_updates = {}
        self._connected = False
        self._disconnected = True

    def _disconnect_plot(self, name):
        plot = self._plots[name]
        plot.disconnect()

    def sync(self):
        """Syncs all connected plots with the session's current view."""
        if not self.is_connected:
            return

        self._update_ids_from_session()
        self._update_all_plots()

    def freeze(self):
        """Freezes all connected plots, replacing them with static images.

        Only applicable in notebook contexts.
        """
        if not self.is_connected:
            return

        for plot in self._plots.values():
            if plot.is_connected:
                plot.freeze()

        self.disconnect()

    def _on_session_update(self, _):
        if not self.is_connected:
            return

        if not self._ready_for_update("session"):
            return

        if self._session.view == self._last_session_view:
            return

        self._update_ids_from_session()
        self._update_plots_from_session()

    def _on_plot_selection(self, name, ids):
        plot = self._plots[name]

        if not plot.is_connected:
            return

        if not self._ready_for_update(name):
            return

        if plot.init_view is not None:
            plot_view = plot.init_view.view()
        else:
            plot_view = self._session.dataset.view()

        sample_ids = None
        frame_ids = None
        labels = None

        if ids is None:
            # Plot is in default state
            pass
        elif plot.link_type == "samples":
            # Update `plot_view` to only contain the right content
            if isinstance(
                plot_view, (fop.PatchesView, fov.FramesView, foc.ClipsView)
            ):
                # Create a view that only contains the selected samples
                plot_view = plot_view.select_by("sample_id", ids)
            else:
                # Create a view that only contains the selected samples
                plot_view = plot_view.select(ids)

            # This plot is linked to samples, so we already know exactly which
            # IDs to use
            sample_ids = ids

            # If the session has plots linked to frames, retrieve the current
            # frame IDs
            if self.has_frame_links:
                if isinstance(plot_view, fov.FramesView):
                    frame_ids = plot_view.values("id")
                elif plot_view._contains_videos():
                    frame_ids = plot_view.values("frames.id", unwind=True)

            # If the session has plots linked to labels, retrieve the current
            # label IDs
            if self.has_label_links:
                labels = plot_view._get_selected_labels()
        elif plot.link_type == "frames":
            # Update `plot_view` to only contain the right content
            if isinstance(plot_view, fov.FramesView):
                # Create a view that only contains the selected frames
                plot_view = plot_view.select(ids)
            elif plot.selection_mode == "select":
                # Create a view that only contains the selected frames
                plot_view = plot_view.select_frames(ids)
            elif plot.selection_mode == "match":
                # Create a view that only contains unfiltered samples with at
                # least one selected frame
                _ids = [ObjectId(_id) for _id in ids]
                plot_view = plot_view.match(F("frames._id").contains(_ids))
            elif plot.selection_mode == "frames":
                # We shouldn't actually get here, since `plot_view` should have
                # been a `FramesView`
                _ids = [ObjectId(_id) for _id in ids]
                plot_view = plot_view.match(F("frames._id").contains(_ids))
            else:
                raise ValueError(
                    "Unsupported `selection_mode=%s`" % plot.selection_mode
                )

            # If the session has plots linked to samples, retrieve the current
            # sample IDs
            if self.has_sample_links:
                if isinstance(plot_view, fov.FramesView):
                    sample_ids = plot_view.values("sample_id")
                else:
                    sample_ids = plot_view.values("id")

            # This plot is linked to frames, so we already know exactly which
            # IDs to use
            frame_ids = ids

            # If the session has plots linked to labels, retrieve the current
            # label IDs
            if self.has_label_links:
                labels = plot_view._get_selected_labels()
        elif plot.link_type == "labels":
            # Update `plot_view` to only contain the right content
            if plot.selection_mode == "select":
                # Create a view that only contains the selected labels
                plot_view = plot_view.select_labels(
                    ids=ids, fields=plot.label_fields
                )
            elif plot.selection_mode == "match":
                # Create a view that only contains unfiltered samples with at
                # least one selected label
                plot_view = plot_view.match_labels(
                    ids=ids, fields=plot.label_fields
                )
            elif plot.selection_mode == "patches":
                if isinstance(plot_view, fop.PatchesView):
                    # Create a view that only contains the selected labels
                    plot_view = plot_view.select(ids)
                else:
                    # We shouldn't actually get here, since `plot_view` should
                    # have been a `PatchesView`...
                    plot_view = plot_view.match_labels(
                        ids=ids, fields=plot.label_fields
                    )
            else:
                raise ValueError(
                    "Unsupported `selection_mode=%s`" % plot.selection_mode
                )

            field = plot.label_fields
            if field is not None and not etau.is_str(field):
                field = None  # multiple fields; unclear which one to use

            # If the session has plots linked to samples, retrieve the current
            # sample IDs
            if self.has_sample_links:
                if isinstance(plot_view, fov.FramesView):
                    sample_ids = plot_view.values("sample_id")
                else:
                    sample_ids = plot_view.values("id")

            # If the session has plots linked to frames, retrieve the current
            # frame IDs
            if self.has_frame_links:
                if isinstance(plot_view, fov.FramesView):
                    frame_ids = plot_view.values("id")
                elif plot_view._contains_videos():
                    frame_ids = plot_view.values("frames.id", unwind=True)

            # This plot is linked to labels, so we already know exactly which
            # IDs to use
            labels = [{"field": field, "label_id": _id} for _id in ids]
        else:
            raise ValueError(
                "Plot '%s' has unsupported link type '%s'"
                % (name, plot.link_type)
            )

        self._current_sample_ids = sample_ids
        self._current_frame_ids = frame_ids
        self._current_labels = labels
        self._update_session(plot_view)
        self._update_plots_from_session(exclude=[name])

    def _update_ids_from_session(self):
        session = self._session
        current_view = session._collection.view()
        has_view = session.view is not None
        has_selections = session.selected or session.selected_labels

        # If no view is loaded and nothing is selected, reset all plots
        if not has_view and not has_selections:
            self._current_sample_ids = None
            self._current_frame_ids = None
            self._current_labels = None
            return

        if self.has_sample_links:
            # If samples are selected in the App, only record those
            # Otherwise, record all samples in the view
            if session.selected:
                _view = current_view.select(session.selected)
            else:
                _view = current_view

            if isinstance(
                _view, (fop.PatchesView, fov.FramesView, foc.ClipsView)
            ):
                sample_ids = _view.values("sample_id")
            else:
                sample_ids = _view.values("id")

            self._current_sample_ids = sample_ids

        if self.has_frame_links:
            # If samples are selected in the App, only record their frame IDs
            # Otherwise, record all frame IDs in the view
            if session.selected:
                _view = current_view.select(session.selected)
            else:
                _view = current_view

            if isinstance(_view, fov.FramesView):
                frame_ids = _view.values("id")
            elif _view._contains_videos():
                frame_ids = _view.values("frames.id", unwind=True)
            else:
                frame_ids = None

            self._current_frame_ids = frame_ids

        if self.has_label_links:
            # If labels are selected in the App, only record those
            # If samples are selected in the App, only record their labels
            # Otherwise, record all labels in the current view
            if session.selected_labels:
                labels = session.selected_labels
            else:
                if session.selected:
                    _view = current_view.select(session.selected)
                else:
                    _view = current_view

                labels = _view._get_selected_labels()

            self._current_labels = labels

    def _update_session(self, view):
        if not self._needs_update("session"):
            return

        self._last_session_view = view
        with self._session.no_show():
            self._session.view = view

    def _update_plots_from_session(self, exclude=None):
        exclude = set(exclude or [])
        names = [
            name
            for name, plot in self._plots.items()
            if plot.supports_session_updates and name not in exclude
        ]
        self._update_plots(names)

    def _update_all_plots(self):
        names = list(self._plots.keys())
        self._update_plots(names)

    def _update_plots(self, names):
        view_plot_names = []
        interactive_plot_names = []
        for name in names:
            plot = self._plots[name]

            if not plot.is_connected:
                continue

            if not self._needs_update(name):
                continue

            if plot.link_type == "view":
                view_plot_names.append(name)
            elif plot.link_type in ("samples", "frames", "labels"):
                interactive_plot_names.append(name)
            else:
                raise ValueError(
                    "Plot '%s' has unsupported link type '%s'"
                    % (name, plot.link_type)
                )

        view = self._session._collection.view()

        if view_plot_names:
            self._update_view_plots(view_plot_names, view)

        for name in interactive_plot_names:
            self._update_interactive_plot(name, view)

    def _update_view_plots(self, names, view):
        # For efficiency, aggregations for all supported `ViewPlot`s are
        # computed in a single batch

        # Build flat list of all aggregations
        aggregations = []
        agg_lengths = []
        for name in names:
            aggs = self._aggs.get(name, None)
            if aggs is not None:
                aggregations.extend(aggs)
                agg_lengths.append(len(aggs))
            else:
                agg_lengths.append(None)

        if aggregations:
            # Compute aggregations
            agg_results = view.aggregate(aggregations)

            # Partition into per-plot batches
            agg_iter = iter(agg_results)
            agg_results = []
            for agg_len in agg_lengths:
                if agg_len is None:
                    _agg_results = None
                else:
                    _agg_results = [next(agg_iter) for _ in range(agg_len)]

                agg_results.append(_agg_results)
        else:
            agg_results = itertools.repeat(None)

        # Finally, update the plots
        for name, _agg_results in zip(names, agg_results):
            plot = self._plots[name]
            plot.update_view(view, agg_results=_agg_results)

    def _update_interactive_plot(self, name, view):
        plot = self._plots[name]

        if plot.link_type == "samples":
            plot.select_ids(self._current_sample_ids, view=view)
        elif plot.link_type == "frames":
            plot.select_ids(self._current_frame_ids, view=view)
        elif plot.link_type == "labels":
            label_ids = self._get_current_label_ids_for_plot(plot)
            plot.select_ids(label_ids, view=view)
        else:
            raise ValueError(
                "InteractivePlot '%s' has unsupported link type '%s'"
                % (name, plot.link_type)
            )

    def _get_current_label_ids_for_plot(self, plot):
        if self._current_labels is None:
            return None

        label_fields = plot.label_fields

        if not isinstance(label_fields, (list, set, tuple)):
            label_fields = [label_fields]

        return [
            l["label_id"]
            for l in self._current_labels
            if l["field"] in label_fields
        ]

    def _ready_for_update(self, name):
        now = datetime.datetime.utcnow()

        if self._last_update is None:
            is_new_update = True
        else:
            delta = (now - self._last_update).total_seconds()
            is_new_update = delta > self._MIN_UPDATE_DELTA_SECONDS

        if is_new_update:
            self._last_update = now
            self._last_updates[name] = now

        return is_new_update

    def _needs_update(self, name):
        now = datetime.datetime.utcnow()

        last_update = self._last_updates.get(name, None)

        if last_update is None:
            ready = True
        else:
            delta = (now - last_update).total_seconds()
            ready = delta > self._MIN_UPDATE_DELTA_SECONDS

        if ready:
            self._last_updates[name] = now

        return ready
